# -*- coding: utf-8 -*-
#
# AWL simulator - Block tree widget
#
# Copyright 2015-2016 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.gui.util import *
from awlsim.gui.icons import *


class BlockTreeModel(QAbstractItemModel):
	EnumGen.start
	INDEXID_SRCS			= EnumGen.item
	INDEXID_SRCS_AWL		= EnumGen.item
	INDEXID_SRCS_SYMTAB		= EnumGen.item
	INDEXID_SRCS_LIBSEL		= EnumGen.item
	INDEXID_BLOCKS			= EnumGen.item
	INDEXID_BLOCKS_OBS		= EnumGen.item
	INDEXID_BLOCKS_FCS		= EnumGen.item
	INDEXID_BLOCKS_FBS		= EnumGen.item
	INDEXID_BLOCKS_DBS		= EnumGen.item
	INDEXID_HWMODS			= EnumGen.item
	EnumGen.end

	# Base ID mask for dynamic elements
	INDEXID_BASE_MASK		= 0xFF0000
	# ID bases for dynamic elements
	INDEXID_SRCS_AWL_BASE		= 0x010000
	INDEXID_SRCS_SYMTAB_BASE	= 0x020000
	INDEXID_SRCS_LIBSEL_BASE	= 0x030000
	INDEXID_BLOCKS_OBS_BASE		= 0x040000
	INDEXID_BLOCKS_FCS_BASE		= 0x050000
	INDEXID_BLOCKS_FBS_BASE		= 0x060000
	INDEXID_BLOCKS_DBS_BASE		= 0x070000
	INDEXID_HWMODS_BASE		= 0x080000

	row2id_toplevel = {
		0	: INDEXID_SRCS,
		1	: INDEXID_BLOCKS,
		2	: INDEXID_HWMODS,
	}
	id2row_toplevel = pivotDict(row2id_toplevel)

	row2id_srcs = {
		0	: INDEXID_SRCS_AWL,
		1	: INDEXID_SRCS_SYMTAB,
		2	: INDEXID_SRCS_LIBSEL,
	}
	id2row_srcs = pivotDict(row2id_srcs)

	row2id_blocks = {
		0	: INDEXID_BLOCKS_OBS,
		1	: INDEXID_BLOCKS_FCS,
		2	: INDEXID_BLOCKS_FBS,
		3	: INDEXID_BLOCKS_DBS,
	}
	id2row_blocks = pivotDict(row2id_blocks)

	EnumGen.start
	COLUMN_NAME	= EnumGen.item
	COLUMN_DESC	= EnumGen.item
	COLUMN_IDENT	= EnumGen.item
	EnumGen.end

	def __init__(self, client, parent=None):
		QAbstractItemModel.__init__(self, parent)
		self.client = client

		self.__awlSources = []		# List of AwlSource()s
		self.__symTabSources = []	# List of SymTabSource()s
		self.__symTabInfoList = []	# List of (SymTabSource(), SymbolTable())
		self.__libSelections = []	# List of AwlLibEntrySelection()s
		self.__obInfos = []		# List of BlockInfo()s for OBs
		self.__fcInfos = []		# List of BlockInfo()s for FCs
		self.__fbInfos = []		# List of BlockInfo()s for FBs
		self.__dbInfos = []		# List of BlockInfo()s for DBs
		self.__hwMods = []		# List of HwmodDescriptor()s

	def getAwlSource(self, indexNr):
		return self.__awlSources[indexNr]

	def getSymTabSource(self, indexNr):
		return self.__symTabSources[indexNr]

	def getOBBlockInfo(self, indexNr):
		return self.__obInfos[indexNr]

	def getFCBlockInfo(self, indexNr):
		return self.__fcInfos[indexNr]

	def getFBBlockInfo(self, indexNr):
		return self.__fbInfos[indexNr]

	def getDBBlockInfo(self, indexNr):
		return self.__dbInfos[indexNr]

	def __getSymbolsByType(self, dataType):
		return itertools.chain.from_iterable(
			symTab.getByDataType(dataType)
			for symTabSrc, symTab in self.__symTabInfoList
		)

	def __getSymbolsByTypeName(self, typeNameString):
		dataType = AwlDataType.makeByName(typeNameString.split())
		return self.__getSymbolsByType(dataType)

	def __updateData(self, localList, newList, parentIndex):
		for i, newItem in enumerate(newList):
			# Add new items.
			if i >= len(localList):
				self.beginInsertRows(parentIndex, i, i)
				localList.append(newItem)
				self.endInsertRows()
				continue
			# Change modified items.
			if newItem != localList[i]:
				self.beginRemoveRows(parentIndex, i, i)
				localList.pop(i)
				self.endRemoveRows()
				self.beginInsertRows(parentIndex, i, i)
				localList.insert(i, newItem)
				self.endInsertRows()
				continue
		while len(localList) > len(newList) and\
		      len(localList) >= 1:
			# Remove removed items.
			i = len(localList) - 1
			self.beginRemoveRows(parentIndex, i, i)
			localList.pop(i)
			self.endRemoveRows()

	def handle_IDENTS(self, msg):
		self.__updateData(self.__awlSources, msg.awlSources,
				  self.idToIndex(self.INDEXID_SRCS_AWL))
		self.__updateData(self.__symTabSources, msg.symTabSources,
				  self.idToIndex(self.INDEXID_SRCS_SYMTAB))
		self.__updateData(self.__libSelections, msg.libSelections,
				  self.idToIndex(self.INDEXID_SRCS_LIBSEL))
		self.__updateData(self.__hwMods, msg.hwMods,
				  self.idToIndex(self.INDEXID_HWMODS))

	def handle_BLOCKINFO(self, msg):
		newBlockInfos = [ bi for bi in msg.blockInfos \
				  if bi.blockType == bi.TYPE_OB ]
		self.__updateData(self.__obInfos, newBlockInfos,
				  self.idToIndex(self.INDEXID_BLOCKS_OBS))

		newBlockInfos = [ bi for bi in msg.blockInfos \
				  if bi.blockType == bi.TYPE_FC ]
		self.__updateData(self.__fcInfos, newBlockInfos,
				  self.idToIndex(self.INDEXID_BLOCKS_FCS))

		newBlockInfos = [ bi for bi in msg.blockInfos \
				  if bi.blockType == bi.TYPE_FB ]
		self.__updateData(self.__fbInfos, newBlockInfos,
				  self.idToIndex(self.INDEXID_BLOCKS_FBS))

		newBlockInfos = [ bi for bi in msg.blockInfos \
				  if bi.blockType == bi.TYPE_DB ]
		self.__updateData(self.__dbInfos, newBlockInfos,
				  self.idToIndex(self.INDEXID_BLOCKS_DBS))

	def handle_symTabInfo(self, symTabInfoList):
		self.__symTabInfoList = symTabInfoList
		#TODO we probably need to trigger row update here.

	def idToIndex(self, idxId, column = 0):
		for table in (self.id2row_toplevel,
			      self.id2row_srcs,
			      self.id2row_blocks):
			if idxId in table:
				return self.createIndex(table[idxId],
							column, idxId)
		assert(0)

	def indexToId(self, index):
		return index.internalId()

	def flags(self, index):
		if not index.isValid():
			return Qt.NoItemFlags

		idxId = self.indexToId(index)
		idxIdBase = idxId & self.INDEXID_BASE_MASK

		if index.column() == self.COLUMN_IDENT:
			if idxIdBase != 0:
				# We set the 'editable' flag for the ident column.
				# However it not really is editable (see setData).
				# We just do this to make the ident easily copy-able.
				return Qt.ItemIsEnabled |\
				       Qt.ItemIsSelectable |\
				       Qt.ItemIsEditable

		return Qt.ItemIsEnabled |\
		       Qt.ItemIsSelectable

	def columnCount(self, parentIndex=QModelIndex()):
		return 3

	def rowCount(self, parentIndex=QModelIndex()):
		if parentIndex.isValid():
			parentId = self.indexToId(parentIndex)
			if parentId == self.INDEXID_SRCS:
				return len(self.row2id_srcs)
			elif parentId == self.INDEXID_SRCS_AWL:
				return len(self.__awlSources)
			elif parentId == self.INDEXID_SRCS_SYMTAB:
				return len(self.__symTabSources)
			elif parentId == self.INDEXID_SRCS_LIBSEL:
				return len(self.__libSelections)
			elif parentId == self.INDEXID_BLOCKS:
				return len(self.row2id_blocks)
			elif parentId == self.INDEXID_BLOCKS_OBS:
				return len(self.__obInfos)
			elif parentId == self.INDEXID_BLOCKS_FCS:
				return len(self.__fcInfos)
			elif parentId == self.INDEXID_BLOCKS_FBS:
				return len(self.__fbInfos)
			elif parentId == self.INDEXID_BLOCKS_DBS:
				return len(self.__dbInfos)
			elif parentId == self.INDEXID_HWMODS:
				return len(self.__hwMods)
		else:
			return len(self.row2id_toplevel)
		return 0

	def index(self, row, column, parentIndex=QModelIndex()):
		if row < 0 or column < 0:
			return QModelIndex()
		if parentIndex.isValid():
			parentId = self.indexToId(parentIndex)
			if parentId == self.INDEXID_SRCS:
				row2idTable = self.row2id_srcs
			elif parentId == self.INDEXID_SRCS_AWL:
				return self.createIndex(row, column,
					self.INDEXID_SRCS_AWL_BASE + row)
			elif parentId == self.INDEXID_SRCS_SYMTAB:
				return self.createIndex(row, column,
					self.INDEXID_SRCS_SYMTAB_BASE + row)
			elif parentId == self.INDEXID_SRCS_LIBSEL:
				return self.createIndex(row, column,
					self.INDEXID_SRCS_LIBSEL_BASE + row)
			elif parentId == self.INDEXID_BLOCKS:
				row2idTable = self.row2id_blocks
			elif parentId == self.INDEXID_BLOCKS_OBS:
				return self.createIndex(row, column,
					self.INDEXID_BLOCKS_OBS_BASE + row)
			elif parentId == self.INDEXID_BLOCKS_FCS:
				return self.createIndex(row, column,
					self.INDEXID_BLOCKS_FCS_BASE + row)
			elif parentId == self.INDEXID_BLOCKS_FBS:
				return self.createIndex(row, column,
					self.INDEXID_BLOCKS_FBS_BASE + row)
			elif parentId == self.INDEXID_BLOCKS_DBS:
				return self.createIndex(row, column,
					self.INDEXID_BLOCKS_DBS_BASE + row)
			elif parentId == self.INDEXID_HWMODS:
				return self.createIndex(row, column,
					self.INDEXID_HWMODS_BASE + row)
			else:
				return QModelIndex()
		else:
			row2idTable = self.row2id_toplevel
		try:
			idxId = row2idTable[row]
		except KeyError as e:
			return QModelIndex()
		return self.createIndex(row, column, idxId)

	def parent(self, index):
		if not index.isValid():
			return QModelIndex()
		idxId = self.indexToId(index)
		idxIdBase = idxId & self.INDEXID_BASE_MASK
		if idxId in self.id2row_toplevel:
			return QModelIndex()
		elif idxId in self.id2row_srcs:
			return self.idToIndex(self.INDEXID_SRCS)
		elif idxId in self.id2row_blocks:
			return self.idToIndex(self.INDEXID_BLOCKS)
		elif idxIdBase == self.INDEXID_SRCS_AWL_BASE:
			return self.idToIndex(self.INDEXID_SRCS_AWL)
		elif idxIdBase == self.INDEXID_SRCS_SYMTAB_BASE:
			return self.idToIndex(self.INDEXID_SRCS_SYMTAB)
		elif idxIdBase == self.INDEXID_SRCS_LIBSEL_BASE:
			return self.idToIndex(self.INDEXID_SRCS_LIBSEL)
		elif idxIdBase == self.INDEXID_BLOCKS_OBS_BASE:
			return self.idToIndex(self.INDEXID_BLOCKS_OBS)
		elif idxIdBase == self.INDEXID_BLOCKS_FCS_BASE:
			return self.idToIndex(self.INDEXID_BLOCKS_FCS)
		elif idxIdBase == self.INDEXID_BLOCKS_FBS_BASE:
			return self.idToIndex(self.INDEXID_BLOCKS_FBS)
		elif idxIdBase == self.INDEXID_BLOCKS_DBS_BASE:
			return self.idToIndex(self.INDEXID_BLOCKS_DBS)
		elif idxIdBase == self.INDEXID_HWMODS_BASE:
			return self.idToIndex(self.INDEXID_HWMODS)
		return QModelIndex()

	def __data_columnName(self, index, idxId, idxIdBase):
		if idxIdBase == self.INDEXID_SRCS_AWL_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__awlSources):
				return None
			return self.__awlSources[index].name
		elif idxIdBase == self.INDEXID_SRCS_SYMTAB_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__symTabSources):
				return None
			return self.__symTabSources[index].name
		elif idxIdBase == self.INDEXID_SRCS_LIBSEL_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__libSelections):
				return None
			return str(self.__libSelections[index])
		elif idxIdBase == self.INDEXID_BLOCKS_OBS_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__obInfos):
				return None
			return self.__obInfos[index].blockName
		elif idxIdBase == self.INDEXID_BLOCKS_FCS_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__fcInfos):
				return None
			return self.__fcInfos[index].blockName
		elif idxIdBase == self.INDEXID_BLOCKS_FBS_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__fbInfos):
				return None
			return self.__fbInfos[index].blockName
		elif idxIdBase == self.INDEXID_BLOCKS_DBS_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__dbInfos):
				return None
			return self.__dbInfos[index].blockName
		elif idxIdBase == self.INDEXID_HWMODS_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__hwMods):
				return None
			return self.__hwMods[index].getModuleName()

		names = {
		  self.INDEXID_SRCS		: "Sources",
		  self.INDEXID_SRCS_AWL		: "AWL/STL",
		  self.INDEXID_SRCS_SYMTAB	: "Symbols",
		  self.INDEXID_SRCS_LIBSEL	: "Libraries",
		  self.INDEXID_BLOCKS		: "Blocks",
		  self.INDEXID_BLOCKS_OBS	: "OBs",
		  self.INDEXID_BLOCKS_FCS	: "FCs",
		  self.INDEXID_BLOCKS_FBS	: "FBs",
		  self.INDEXID_BLOCKS_DBS	: "DBs",
		  self.INDEXID_HWMODS		: "Hardware",
		}
		return names.get(idxId)

	def __tryGetBlockSymName(self, blockName):
		syms = list(self.__getSymbolsByTypeName(blockName))
		if syms:
			return '"%s"' % syms[-1].getName()
		return None

	def __data_columnDesc(self, index, idxId, idxIdBase):
		if idxIdBase == self.INDEXID_BLOCKS_OBS_BASE:
			obInfoIndex = idxId - idxIdBase
			descs = {
				1	: "Main cycle: \"CYCL_EXC\"",
				100	: "Warm start: \"COMPLETE RESTART\"",
				101	: "Restart: \"RESTART\"",
				102	: "Cold restart: \"COLD RESTART\"",
			}
			obInfo = self.__obInfos[obInfoIndex]
			desc = self.__tryGetBlockSymName("OB %d" % obInfo.blockIndex)
			return desc or descs.get(obInfo.blockIndex)
		elif idxIdBase == self.INDEXID_BLOCKS_FCS_BASE:
			fcInfoIndex = idxId - idxIdBase
			fcInfo = self.__fcInfos[fcInfoIndex]
			return self.__tryGetBlockSymName("FC %d" % fcInfo.blockIndex)
		elif idxIdBase == self.INDEXID_BLOCKS_FBS_BASE:
			fbInfoIndex = idxId - idxIdBase
			fbInfo = self.__fbInfos[fbInfoIndex]
			return self.__tryGetBlockSymName("FB %d" % fbInfo.blockIndex)
		elif idxIdBase == self.INDEXID_BLOCKS_DBS_BASE:
			dbInfoIndex = idxId - idxIdBase
			descs = {
				0 : "System data block",
			}
			dbInfo = self.__dbInfos[dbInfoIndex]
			desc = self.__tryGetBlockSymName("DB %d" % dbInfo.blockIndex)
			return desc or descs.get(dbInfo.blockIndex)
		descs = {
		  self.INDEXID_SRCS		: "Source files",
		  self.INDEXID_SRCS_AWL		: "AWL/STL sources",
		  self.INDEXID_SRCS_SYMTAB	: "Symbol table sources",
		  self.INDEXID_SRCS_LIBSEL	: "Library selections",
		  self.INDEXID_BLOCKS		: "Compiled blocks",
		  self.INDEXID_BLOCKS_OBS	: "Organization Blocks",
		  self.INDEXID_BLOCKS_FCS	: "Functions",
		  self.INDEXID_BLOCKS_FBS	: "Function Blocks",
		  self.INDEXID_BLOCKS_DBS	: "Data Blocks",
		  self.INDEXID_HWMODS		: "Hardware modules",
		}
		return descs.get(idxId)

	def __data_columnIdent(self, index, idxId, idxIdBase):
		if idxIdBase == self.INDEXID_SRCS_AWL_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__awlSources):
				return None
			return self.__awlSources[index].identHashStr
		elif idxIdBase == self.INDEXID_SRCS_SYMTAB_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__symTabSources):
				return None
			return self.__symTabSources[index].identHashStr
		elif idxIdBase == self.INDEXID_SRCS_LIBSEL_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__libSelections):
				return None
			return self.__libSelections[index].getIdentHashStr()
		elif idxIdBase == self.INDEXID_BLOCKS_OBS_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__obInfos):
				return None
			return self.__obInfos[index].identHashStr
		elif idxIdBase == self.INDEXID_BLOCKS_FCS_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__fcInfos):
				return None
			return self.__fcInfos[index].identHashStr
		elif idxIdBase == self.INDEXID_BLOCKS_FBS_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__fbInfos):
				return None
			return self.__fbInfos[index].identHashStr
		elif idxIdBase == self.INDEXID_BLOCKS_DBS_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__dbInfos):
				return None
			return self.__dbInfos[index].identHashStr
		elif idxIdBase == self.INDEXID_HWMODS_BASE:
			index = idxId - idxIdBase
			if index >= len(self.__hwMods):
				return None
			return self.__hwMods[index].getIdentHashStr()
		return None

	def data(self, index, role=Qt.DisplayRole):
		column = index.column()
		idxId = self.indexToId(index)
		idxIdBase = idxId & self.INDEXID_BASE_MASK

		if role in (Qt.DisplayRole, Qt.EditRole):
			if column == self.COLUMN_NAME:
				return self.__data_columnName(index, idxId, idxIdBase)
			if column == self.COLUMN_DESC:
				return self.__data_columnDesc(index, idxId, idxIdBase)
			if column == self.COLUMN_IDENT:
				return self.__data_columnIdent(index, idxId, idxIdBase)

		elif role == Qt.DecorationRole:
			if column == self.COLUMN_NAME:
				if idxId == self.INDEXID_SRCS:
					return getIcon("textsource")
				elif idxId == self.INDEXID_SRCS_AWL or\
				     idxIdBase == self.INDEXID_SRCS_AWL_BASE:
					return getIcon("textsource")
				elif idxId == self.INDEXID_SRCS_SYMTAB or\
				     idxIdBase == self.INDEXID_SRCS_SYMTAB_BASE:
					return getIcon("tag")
				elif idxId == self.INDEXID_SRCS_LIBSEL or\
				     idxIdBase == self.INDEXID_SRCS_LIBSEL_BASE:
					return getIcon("stdlib")
				elif idxId == self.INDEXID_BLOCKS:
					return getIcon("plugin")
				elif idxId == self.INDEXID_BLOCKS_OBS or\
				     idxIdBase == self.INDEXID_BLOCKS_OBS_BASE:
					return getIcon("plugin")
				elif idxId == self.INDEXID_BLOCKS_FCS or\
				     idxIdBase == self.INDEXID_BLOCKS_FCS_BASE:
					return getIcon("plugin")
				elif idxId == self.INDEXID_BLOCKS_FBS or\
				     idxIdBase == self.INDEXID_BLOCKS_FBS_BASE:
					return getIcon("plugin")
				elif idxId == self.INDEXID_BLOCKS_DBS or\
				     idxIdBase == self.INDEXID_BLOCKS_DBS_BASE:
					return getIcon("datablock")
				elif idxId == self.INDEXID_HWMODS or\
				     idxIdBase == self.INDEXID_HWMODS_BASE:
					return getIcon("hwmod")
		return None

	def setData(self, index, value, role=Qt.EditRole):
		if not index.isValid():
			return False
		if role != Qt.EditRole:
			return False
		if index.column() == self.COLUMN_IDENT:
			# We never allow edit of the ident hash.
			return False
		return QAbstractItemModel.setData(self, index, value, role)

	def headerData(self, section, orientation, role=Qt.DisplayRole):
		if role == Qt.DisplayRole:
			return (
				"Block name",
				"Symbol / description",
				"Unique identification hash",
			)[section]
		return None

class BlockTreeView(QTreeView):
	def __init__(self, model, parent=None):
		QTreeView.__init__(self, parent)
		self.setModel(model)
		if model:
			self.expand(model.idToIndex(model.INDEXID_SRCS))
			self.expand(model.idToIndex(model.INDEXID_BLOCKS))
		self.setColumnWidth(0, 170)
		self.setColumnWidth(1, 220)
		self.setColumnWidth(2, 530)

		self.__currentIdxId = None

		self.__blockMenu = QMenu(self)
		self.__blockMenu.addAction("&Remove block from CPU...",
					   self.__removeBlock)

		self.__srcMenu = QMenu(self)
		self.__srcMenu.addAction("&Remove source from CPU...",
					 self.__removeSource)

		self.pressed.connect(self.__mouseBtnPressed)

	def __mouseBtnPressed(self, index):
		buttons = QApplication.mouseButtons()
		model = self.model()
		if not model:
			return

		try:
			idxId = model.indexToId(index)
			idxIdBase = idxId & model.INDEXID_BASE_MASK
			self.__currentIdxId = idxId

			if buttons & Qt.RightButton:
				if idxIdBase == model.INDEXID_BLOCKS_OBS_BASE or\
				   idxIdBase == model.INDEXID_BLOCKS_FCS_BASE or\
				   idxIdBase == model.INDEXID_BLOCKS_FBS_BASE or\
				   idxIdBase == model.INDEXID_BLOCKS_DBS_BASE:
					self.__blockMenu.exec_(QCursor.pos())
				elif idxIdBase == model.INDEXID_SRCS_AWL_BASE or\
				     idxIdBase == model.INDEXID_SRCS_SYMTAB_BASE:
					self.__srcMenu.exec_(QCursor.pos())
		finally:
			self.__currentIdxId = None

	def keyPressEvent(self, ev):
		QTreeView.keyPressEvent(self, ev)

		model = self.model()
		if not model:
			return
		idxId = model.indexToId(self.currentIndex())
		idxIdBase = idxId & model.INDEXID_BASE_MASK

		if ev.key() == Qt.Key_Delete:
			if idxIdBase == model.INDEXID_BLOCKS_OBS_BASE or\
			   idxIdBase == model.INDEXID_BLOCKS_FCS_BASE or\
			   idxIdBase == model.INDEXID_BLOCKS_FBS_BASE or\
			   idxIdBase == model.INDEXID_BLOCKS_DBS_BASE:
				self.__removeBlock()
			if idxIdBase == model.INDEXID_SRCS_AWL_BASE or\
			   idxIdBase == model.INDEXID_SRCS_SYMTAB_BASE:
				self.__removeSource()

	def __removeSource(self):
		model = self.model()
		if not model:
			return
		client = model.client

		res = QMessageBox.question(self,
			"Remove selected source?",
			"Remove the selected source from the CPU?\n"
			"This will also remove all associated compiled "
			"blocks from the CPU.",
			QMessageBox.Yes | QMessageBox.No,
			QMessageBox.Yes)
		if res != QMessageBox.Yes:
			return

		if self.__currentIdxId is None:
			idxId = model.indexToId(self.currentIndex())
		else:
			idxId = self.__currentIdxId
		idxIdBase = idxId & model.INDEXID_BASE_MASK
		indexNr = idxId - idxIdBase

		if idxIdBase == model.INDEXID_SRCS_AWL_BASE:
			identHash = model.getAwlSource(indexNr).identHash
		elif idxIdBase == model.INDEXID_SRCS_SYMTAB_BASE:
			identHash = model.getSymTabSource(indexNr).identHash
		else:
			return
		try:
			client.removeSource(identHash)
		except AwlSimError as e:
			MessageBox.handleAwlSimError(self,
				"An error occurred while removing a source", e)

	def __removeBlock(self):
		model = self.model()
		if not model:
			return
		client = model.client

		res = QMessageBox.question(self,
			"Remove selected block?",
			"Remove the selected block from the CPU?",
			QMessageBox.Yes | QMessageBox.No,
			QMessageBox.Yes)
		if res != QMessageBox.Yes:
			return

		if self.__currentIdxId is None:
			idxId = model.indexToId(self.currentIndex())
		else:
			idxId = self.__currentIdxId
		idxIdBase = idxId & model.INDEXID_BASE_MASK
		indexNr = idxId - idxIdBase

		if idxIdBase == model.INDEXID_BLOCKS_OBS_BASE:
			blockInfo = model.getOBBlockInfo(indexNr)
		elif idxIdBase == model.INDEXID_BLOCKS_FCS_BASE:
			blockInfo = model.getFCBlockInfo(indexNr)
		elif idxIdBase == model.INDEXID_BLOCKS_FBS_BASE:
			blockInfo = model.getFBBlockInfo(indexNr)
		elif idxIdBase == model.INDEXID_BLOCKS_DBS_BASE:
			blockInfo = model.getDBBlockInfo(indexNr)
		else:
			return
		try:
			client.removeBlock(blockInfo)
		except AwlSimError as e:
			MessageBox.handleAwlSimError(self,
				"An error occurred while removing a block", e)



# -*- coding: utf-8 -*-
#
# AWL simulator - block interface
#
# Copyright 2012-2017 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.common.util import *
from awlsim.common.enumeration import *
from awlsim.common.exceptions import *

from awlsim.core.memory import * #+cimport
from awlsim.core.identifier import *
from awlsim.core.datastructure import * #+cimport
from awlsim.core.operatortypes import * #+cimport
from awlsim.core.operators import * #+cimport


__all__ = [
	"BlockInterfaceField",
	"VerboseBlockIntfField",
	"BlockInterface",
	"OBInterface",
	"FBInterface",
	"FCInterface",
]


class BlockInterfaceField(object):
	"""Block interface field descriptor."""

	EnumGen.start = -1
	FTYPE_UNKNOWN	= EnumGen.item
	FTYPE_IN	= EnumGen.item
	FTYPE_OUT	= EnumGen.item
	FTYPE_INOUT	= EnumGen.item
	FTYPE_STAT	= EnumGen.item
	FTYPE_TEMP	= EnumGen.item
	EnumGen.end

	__slots__ = (
		"name",
		"__dataType",
		"fieldType",
		"fieldIndex",
	)

	def __init__(self, name, dataType):
		# name -> The name string of the field, as defined
		#         in the block interface definition.
		# dataType -> AwlDataType instance or a type name string.
		self.name = name
		self.__dataType = dataType
		self.fieldType = self.FTYPE_UNKNOWN	# set later
		self.fieldIndex = None			# set later

	@property
	def dataType(self):
		from awlsim.core.datatypes import AwlDataType

		dataType = self.__dataType
		if not isinstance(dataType, AwlDataType):
			dataType = AwlDataType.makeByName(dataType)
			self.__dataType = dataType
		return dataType

	def varDeclString(self):
		return "%s : %s;" %\
			(self.name, str(self.dataType))

	def __repr__(self):
		ftype = {
			self.FTYPE_UNKNOWN	: "UNKNOWN",
			self.FTYPE_IN		: "IN",
			self.FTYPE_OUT		: "OUT",
			self.FTYPE_INOUT	: "IN_OUT",
			self.FTYPE_STAT		: "STAT",
			self.FTYPE_TEMP		: "TEMP",
		}[self.fieldType]
		return "(%s)  %s" %\
			(ftype, BlockInterfaceField.varDeclString(self))

class VerboseBlockIntfField(BlockInterfaceField):
	"""Block interface field descriptor,
	with verbose description string."""

	__slots__ = (
		"desc",
	)

	def __init__(self, name, dataType, desc=""):
		BlockInterfaceField.__init__(self, name, dataType)
		self.desc = desc

	def varDeclString(self):
		ret = [ BlockInterfaceField.varDeclString(self), ]
		if self.desc:
			ret.append(" // ")
			ret.append(self.desc)
		return "".join(ret)

class BlockInterface(object):
	"""Code block interface (IN/OUT/IN_OUT/STAT/TEMP parameters) base class."""

	# Structs (self._struct and self.tempStruct) build status.
	EnumGen.start
	STRUCT_NOT_BUILT	= EnumGen.item # Structs not built, yet.
	STRUCT_BUILDING		= EnumGen.item # Currently building structs.
	STRUCT_BUILT		= EnumGen.item # Structs are completely built.
	EnumGen.end

	# Specifies whether this interface has a DI assigned.
	# Set to true for FBs and SFBs
	hasInstanceDB = False

	# The number of allocated bytes on startup.
	# This is only non-zero for OBs
	startupTempAllocation = 0

	class StructRecursion(Exception): pass

	def __init__(self):
		self._struct = None # Instance-DB structure (IN, OUT, INOUT, STAT)
		self.tempStruct = None # Local-stack structure (TEMP)
		self.__structState = self.STRUCT_NOT_BUILT

		self.fieldNameMap = {}
		self.fields_IN = []
		self.fields_OUT = []
		self.fields_INOUT = []
		self.fields_STAT = []
		self.fields_TEMP = []

		self.interfaceFieldCount = 0	# The number of interface fields
		self.staticFieldCount = 0	# The number of static fields
		self.tempFieldCount = 0		# The number of temp fields
		self.tempAllocation = 0		# The number of allocated TEMP bytes (interface only!)

	@property
	def fields_IN_OUT_INOUT(self):
		ret = self.fields_IN[:]
		ret.extend(self.fields_OUT)
		ret.extend(self.fields_INOUT)
		return ret

	@property
	def fields_IN_OUT_INOUT_STAT(self):
		ret = self.fields_IN_OUT_INOUT # don't copy
		ret.extend(self.fields_STAT)
		return ret

	def __addField(self, field):
		if field.name in self.fieldNameMap:
			raise AwlSimError("Data structure field name '%s' is ambiguous." %\
				field.name)

		if field.fieldType == BlockInterfaceField.FTYPE_TEMP:
			self.tempFieldCount += 1
		elif field.fieldType == BlockInterfaceField.FTYPE_STAT:
			self.staticFieldCount += 1
		else:
			self.interfaceFieldCount += 1
		self.fieldNameMap[field.name] = field

	def addField_IN(self, field):
		field.fieldType = field.FTYPE_IN
		self.__addField(field)
		self.fields_IN.append(field)

	def addField_OUT(self, field):
		field.fieldType = field.FTYPE_OUT
		self.__addField(field)
		self.fields_OUT.append(field)

	def addField_INOUT(self, field):
		field.fieldType = field.FTYPE_INOUT
		self.__addField(field)
		self.fields_INOUT.append(field)

	def addField_STAT(self, field):
		field.fieldType = field.FTYPE_STAT
		self.__addField(field)
		self.fields_STAT.append(field)

	def addField_TEMP(self, field):
		field.fieldType = field.FTYPE_TEMP
		self.__addField(field)
		self.fields_TEMP.append(field)

	# Set fieldIndex of each field.
	def __enumerateFields(self):
		for i, field in enumerate(self.fields_IN_OUT_INOUT):
			field.fieldIndex = i

	def __buildField(self, cpu, struct, field, isFirst):
		if isFirst:
			# Align to word boundary.
			structField = struct.addFieldAligned(cpu,
							field.name,
							field.dataType, 2)
		else:
			# Align naturally.
			structField = struct.addFieldNaturallyAligned(cpu,
							field.name,
							field.dataType)
		if self.hasInstanceDB and\
		   field.fieldType == field.FTYPE_INOUT and\
		   field.dataType.compound:
			# This is an FB IN_OUT compound data type parameter.
			# These are special. They are passed via DB-pointer.
			# The DB-pointer is stored in the instance DB.
			# Set the struct field override for this struct field
			# to DB-pointer type.
			structField.override = AwlStructField(
					structField.name, structField.offset,
					"POINTER")

	def __resolveMultiInstanceField(self, cpu, field):
		from awlsim.core.datatypes import AwlDataType

		if field.dataType.type != AwlDataType.TYPE_FB_X and\
		   field.dataType.type != AwlDataType.TYPE_SFB_X:
			# Not a multi-instance field.
			return
		if field.dataType.width >= 0:
			# Already resolved.
			return

		# This is a multi-instance element.
		# Get the FB that is embedded as multi-instance.
		try:
			if field.dataType.type == AwlDataType.TYPE_SFB_X:
				multiFB = cpu.sfbs[field.dataType.index]
			else:
				multiFB = cpu.fbs[field.dataType.index]
		except KeyError:
			raise AwlSimError("The function block '%s' of the "
				"embedded multi-instance '%s' "
				"does not exist." %\
				(str(field.dataType), str(field.name)))

		# This embedded (S)FB might contain multi-instances, too.
		# Resolve them first. Do this by building its data structure.
		# Catch recursions.
		try:
			multiFB.interface.buildDataStructure(cpu)
		except self.StructRecursion:
			raise AwlSimError("Recursion detected while trying to "
				"resolve the multiinstance variables in %s." %\
				(str(multiFB)))

		# Assign the type width, in bits.
		field.dataType.width = multiFB.interface._struct.getSize() * 8
		if field.dataType.width == 0:
			# This is not supported by S7.
			# Awlsim _could_ support it, though.
			raise AwlSimError("Multiinstances with zero size "
				"are not supported. Please declare at least "
				"one static, input, output or in_out variable "
				"in %s" % str(multiFB))

	def buildDataStructure(self, cpu):
		if self.__structState == self.STRUCT_BUILT:
			# We already built this block's interface structure.
			return
		if self.__structState == self.STRUCT_BUILDING:
			# Whoops, we recursed! This is bad.
			raise self.StructRecursion()
		self.__structState = self.STRUCT_BUILDING

		self.__enumerateFields()

		# Build instance-DB structure, if any.
		if self.hasInstanceDB:
			# Resolve the sizes of all multi-instance
			# fields in our STAT area.
			for field in self.fields_STAT:
				self.__resolveMultiInstanceField(cpu, field)

			# Build instance-DB structure for the FB
			self._struct = AwlStruct()
			for i, field in enumerate(self.fields_IN):
				self.__buildField(cpu, self._struct, field, i==0)
			for i, field in enumerate(self.fields_OUT):
				self.__buildField(cpu, self._struct, field, i==0)
			for i, field in enumerate(self.fields_INOUT):
				self.__buildField(cpu, self._struct, field, i==0)
			for i, field in enumerate(self.fields_STAT):
				self.__buildField(cpu, self._struct, field, i==0)
		else:
			# An FC does not have an instance-DB
			assert(not self.fields_STAT) # No static data.

		# Build local-stack structure
		self.tempStruct = AwlStruct()
		for i, field in enumerate(self.fields_TEMP):
			self.__buildField(cpu, self.tempStruct, field, i==0)
		self.tempAllocation = self.tempStruct.getSize()
		# If the OB-interface did not specify all automatic TEMP-fields,
		# just force allocate them, so the lstack-allocator will not be confused.
		if self.tempAllocation < self.startupTempAllocation:
			self.tempAllocation = self.startupTempAllocation

		self.__structState = self.STRUCT_BUILT

	# Get an interface field by name string.
	def getFieldByName(self, name):
		try:
			return self.fieldNameMap[name]
		except KeyError:
			raise AwlSimError("Interface field '%s' does not exist." %\
				name)

	# Get the interface field (BlockInterfaceField) by AwlDataIdentChain.
	# Only the first element of the chain is honored.
	# (Other elements belong to sub-STRUCTs.)
	def getFieldByIdentChain(self, identChain):
		fieldName = identChain[0].name
		field = self.getFieldByName(fieldName)
		return field

	# Get the interface field type (FTYPE_xxx) for a
	# given identifier chain (AwlDataIdentChain).
	def getFieldType(self, identChain):
		return self.getFieldByIdentChain(identChain).fieldType

	# Get the data type (AwlDataType) of an interface field
	# identified by a given ident chain (AwlDataIdentChain).
	# If 'deep' is True, get the data type of the last element
	# in the chain. If 'deep' is False, get the type of the first
	# element (the interface element) only.
	def getFieldDataType(self, identChain, deep=True):
		# The interface field is determined by the first ident
		# chain element.
		fieldName = identChain[0].name
		field = self.getFieldByName(fieldName)
		if len(identChain) == 1 or not deep:
			return field.dataType
		identChain = AwlDataIdentChain(identChain[1:])
		# Walk the rest of the identifier chain to get to
		# the data type of the final element.
		structFieldName = identChain.dup(withIndices=False).getString()
		if field.dataType.itemStruct is None:
			raise AwlSimError("Data type '%s' does not have sub fields. "
				"Resolve of sub field '%s' failed." %\
				(str(field.dataType), structFieldName))
		structField = field.dataType.itemStruct.getField(structFieldName)
		return structField.dataType

	# Get an AwlOperator for TEMP access.
	def __getOperatorForField_TEMP(self, identChain, wantPointer):
		structField = self.tempStruct.getField(
			identChain.dup(withIndices=False).getString())
		if wantPointer:
			ptrValue = structField.offset.toPointerValue()
			ptrValue |= PointerConst.AREA_L_S
			oper = make_AwlOperator(operType=AwlOperatorTypes.IMM_PTR,
					   width=32,
					   offset=None,
					   insn=None)
			oper.pointer = Pointer(ptrValue)
			return oper
		oper = make_AwlOperator(operType=AwlOperatorTypes.MEM_L,
				   width=structField.bitSize,
				   offset=structField.offset.dup(),
				   insn=None)
		# If this is a compound data type access, mark
		# the operand as such.
		oper.compound = structField.dataType.compound
		return oper

	# Get an AwlOperator that addresses the specified interface field identified
	# by "identChain", which is an AwlDataIdentChain instance.
	# If wantPointer is true, an IMM_PTR AwlOperator to the interfaceField
	# is returned.
	def getOperatorForField(self, identChain, wantPointer):
		from awlsim.core.datatypes import AwlDataType

		interfaceFieldType = self.getFieldType(identChain)

		if interfaceFieldType == BlockInterfaceField.FTYPE_TEMP:
			# get TEMP interface field operator
			return self.__getOperatorForField_TEMP(identChain,
							       wantPointer)
		# otherwise get IN/OUT/INOUT/STAT interface field operator

		structField = self._struct.getField(identChain.getString())

		# FC-parameters cannot be resolved statically.
		assert(self.hasInstanceDB)

		if wantPointer:
			ptrValue = structField.offset.toPointerValue()
			ptrValue |= PointerConst.AREA_DI_S
			oper = make_AwlOperator(operType=AwlOperatorTypes.IMM_PTR,
					   width=32,
					   offset=None,
					   insn=None)
			oper.pointer = Pointer(ptrValue)
			return oper

		# Translate to instance-DB access

		if structField.dataType.type in AwlDataType.callByRefTypes:
			# "call by reference"
			offsetOper = make_AwlOperator(operType=AwlOperatorTypes.MEM_DI,
						 width=structField.dataType.width,
						 offset=structField.offset.dup(),
						 insn=None)
			if structField.dataType.type == AwlDataType.TYPE_TIMER:
				area = AwlIndirectOpConst.EXT_AREA_T_S
				width = 16
			elif structField.dataType.type == AwlDataType.TYPE_COUNTER:
				area = AwlIndirectOpConst.EXT_AREA_Z_S
				width = 16
			elif structField.dataType.type == AwlDataType.TYPE_BLOCK_DB:
				area = AwlIndirectOpConst.EXT_AREA_BLKREF_DB_S
				width = 16
			elif structField.dataType.type == AwlDataType.TYPE_BLOCK_FB:
				area = AwlIndirectOpConst.EXT_AREA_BLKREF_FB_S
				width = 16
			elif structField.dataType.type == AwlDataType.TYPE_BLOCK_FC:
				area = AwlIndirectOpConst.EXT_AREA_BLKREF_FC_S
				width = 16
			else:
				assert(0)
				return
			return make_AwlIndirectOp(
				area=area,
				width=width,
				addressRegister=AwlIndirectOpConst.AR_NONE,
				offsetOper=offsetOper,
				insn=None)

		if structField.dataType.type in (AwlDataType.TYPE_FB_X,
						 AwlDataType.TYPE_SFB_X):
			# Multi-instance operator (CALL)
			if structField.dataType.type == AwlDataType.TYPE_FB_X:
				operType = AwlOperatorTypes.MULTI_FB
			else:
				operType = AwlOperatorTypes.MULTI_SFB
			offset = structField.offset.dup()
			offset.fbNumber = structField.dataType.index
			return make_AwlOperator(operType=operType,
					   width=structField.bitSize,
					   offset=offset,
					   insn=None)

		# "call by value"
		oper = make_AwlOperator(operType=AwlOperatorTypes.MEM_DI,
				   width=structField.bitSize,
				   offset=structField.offset.dup(),
				   insn=None)
		# If this is a compound data type access, mark
		# the operand as such.
		oper.compound = structField.dataType.compound
		return oper

	def __repr__(self):
		ret = []
		for flist, fname in ((self.fields_IN, "VAR_IN"),
				     (self.fields_OUT, "VAR_OUT"),
				     (self.fields_INOUT, "VAR_IN_OUT")):
			if not flist:
				continue
			ret.append(fname)
			for field in flist:
				ret.append("  %s : %s;" %\
					   (field.name, str(field.dataType)))
			ret.append("END_VAR")
		if not ret:
			ret = [ "<None>" ]
		return '\n'.join(ret)

class OBInterface(BlockInterface):
	startupTempAllocation = 20

	def addField_IN(self, field):
		raise AwlSimError("VAR_INPUT not possible in an OB")

	def addField_OUT(self, field):
		raise AwlSimError("VAR_OUTPUT not possible in an OB")

	def addField_INOUT(self, field):
		raise AwlSimError("VAR_IN_OUT not possible in an OB")

	def addField_STAT(self, field):
		raise AwlSimError("Static VAR not possible in an OB")

class FBInterface(BlockInterface):
	hasInstanceDB = True

class FCInterface(BlockInterface):
	def addField_STAT(self, field):
		raise AwlSimError("Static VAR not possible in an FC")



# -*- coding: utf-8 -*-
#
# AWL simulator - FUP compiler
#
# Copyright 2016-2017 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.common.sources import *
from awlsim.common.xmlfactory import *
from awlsim.common.cpuconfig import *

from awlsim.core.cpu import * #+cimport
from awlsim.core.operators import * #+cimport
from awlsim.core.operatortypes import * #+cimport
from awlsim.core.statusword import * #+cimport

from awlsim.core.instructions.all_insns import * #+cimport

from awlsim.awloptimizer.awloptimizer import *

from awlsim.fupcompiler.blockdecl import *
from awlsim.fupcompiler.interf import *
from awlsim.fupcompiler.grid import *
from awlsim.fupcompiler.elem import *


class FupFakeCpu(S7CPU):
	def getMnemonics(self):
		return self.fupCompiler.mnemonics

class FupCompilerFactory(XmlFactory):
	FUP_VERSION = 0

	def parser_open(self, tag=None):
		self.inFup = False
		self.inGrids = False
		XmlFactory.parser_open(self, tag)

	def parser_beginTag(self, tag):
		if self.inFup:
			if self.inGrids:
				if tag.name == "grid":
					grid = FupCompiler_Grid(self.compiler)
					self.compiler.grids.append(grid)
					self.parser_switchTo(grid.factory(grid=grid))
					return
			else:
				if tag.name == "blockdecl":
					if self.compiler.decl:
						raise self.Error("Multiple <blockdecl>s defined.")
					decl = FupCompiler_BlockDecl(self.compiler)
					self.compiler.decl = decl
					self.parser_switchTo(decl.factory(decl=decl))
					return
				if tag.name == "interface":
					if self.compiler.interf:
						raise self.Error("Multiple <interface>s defined.")
					interf = FupCompiler_Interf(self.compiler)
					self.compiler.interf = interf
					self.parser_switchTo(interf.factory(interf=interf))
					return
				if tag.name == "grids":
					self.inGrids = True
					return
		else:
			if tag.name == "FUP":
				version = tag.getAttrInt("version")
				if version != self.FUP_VERSION:
					raise self.Error("Unsupported FUP version. "
						"Got %d, but expected %d." % (
						version, self.FUP_VERSION))
				self.inFup = True
				return
		XmlFactory.parser_beginTag(self, tag)

	def parser_endTag(self, tag):
		if self.inFup:
			if self.inGrids:
				if tag.name == "grids":
					self.inGrids = False
					return
			else:
				if tag.name == "FUP":
					self.inFup = False
					self.parser_finish()
					return
		XmlFactory.parser_endTag(self, tag)

class FupCompiler(object):

	# Convert U to UN and vice versa.
	invertedInsnClass = {
		AwlInsn_UN	: AwlInsn_U,
		AwlInsn_UNB	: AwlInsn_UB,
		AwlInsn_ON	: AwlInsn_O,
		AwlInsn_ONB	: AwlInsn_OB,
		AwlInsn_XN	: AwlInsn_X,
		AwlInsn_XNB	: AwlInsn_XB,
	}
	invertedInsnClass.update(pivotDict(invertedInsnClass))

	# Convert U to UB and vice versa.
	branchInsnClass = {
		AwlInsn_UB	: AwlInsn_U,
		AwlInsn_UNB	: AwlInsn_UN,
		AwlInsn_OB	: AwlInsn_O,
		AwlInsn_ONB	: AwlInsn_ON,
		AwlInsn_XB	: AwlInsn_X,
		AwlInsn_XNB	: AwlInsn_XN,
	}
	branchInsnClass.update(pivotDict(branchInsnClass))

	def __init__(self):
		self.reset()

	def reset(self):
		self.mnemonics = None
		self.opTrans = None
		self.decl = None	# FupCompiler_BlockDecl
		self.interf = None	# FupCompiler_Interf
		self.grids = []		# FupCompiler_Grid

		self.blockHeaderAwl = []	# Block declaration header AWL code strings
		self.blockFooterAwl = []	# Block declaration footer AWL code strings
		self.blockInterfAwl = []	# Block interface AWL code strings
		self.instanceDBsAwl = []	# Instance DBs AWL code strings
		self.fupSource = None		# FUP source
		self.awlSource = None		# Compiled AWL source
		self.optimizer = None		# Optimizer instance, if any
		self.__labelCounter = 0		# Current label name counter

	def getAwlSource(self):
		return self.awlSource

	def newLabel(self):
		"""Generate a new block-unique label.
		A label name string is returned.
		The name does not include the final ':' character.
		"""
		labelMax = 0xFFF
		labelCounter = self.__labelCounter
		if labelCounter > labelMax:
			raise FupCompilerError("Out of jump labels. "
				"Cannot create more than %d labels." % (
				labelMax))
		self.__labelCounter += 1
		return "L%03X" % labelCounter

	def newInsn(self, parentFupElem, insnClass, ops=[]):
		"""Create a new instruction instance.
		parentFupElem: The FUP element that creates this insn.
		insnClass: The class that shall be instantiated.
		ops: Optional list of operators.
		"""
		insn = insnClass(cpu=None, ops=ops)
		if parentFupElem:
			insn.commentStr = str(parentFupElem)
		return insn

	def newInsn_JMP(self, parentFupElem, insnClass, labelStr):
		"""Create a new jump instruction instance.
		parentFupElem: The FUP element that creates this insn.
		insnClass: The jump instruction class that shall be instantiated.
		labelStr: The label name string of the jump target.
		"""
		oper = make_AwlOperator(AwlOperatorTypes.LBL_REF, 0, None, None)
		oper.immediateStr = labelStr
		return self.newInsn(parentFupElem=parentFupElem,
				    insnClass=insnClass,
				    ops=[oper])

	def newInsn_NOP(self, parentFupElem, labelStr):
		"""Create a new jump instruction instance.
		parentFupElem: The FUP element that creates this insn.
		labelStr: Optional label name string.
		"""
		oper = make_AwlOperator(AwlOperatorTypes.IMM, 16, None, None)
		oper.immediate = 0
		insn = self.newInsn(parentFupElem, AwlInsn_NOP, ops=[oper])
		if labelStr:
			insn.labelStr = labelStr
		return insn

	def newInsn_LOAD_BIE(self, parentFupElem, insnClass):
		"""Create a new BIE load instruction instance.
		parentFupElem: The FUP element that creates this insn.
		insnClass: The loadinstruction class that shall be instantiated (i.e. U)
		"""
		bitPos = S7StatusWord.getBitnrByName("BIE", S7CPUConfig.MNEMONICS_DE)
		offset = make_AwlOffset(0, bitPos)
		oper = make_AwlOperator(AwlOperatorTypes.MEM_STW, 1, offset, None)
		return self.newInsn(parentFupElem, insnClass, ops=[oper])

	def getOperDataWidth(self, oper):
		"""Helper function to get the data type width (in bits)
		of an operator. This will first attempt to resolve the operator,
		if it is symbolic.
		"""
		from awlsim.core.datatypes import AwlDataType

		if oper.width > 0:
			return oper.width
		if oper.operType == AwlOperatorTypes.NAMED_LOCAL:
			# Get the type of an interface field
			# and return its bit width.
			fieldName = str(oper.offset.identChain)
			field = self.interf.getFieldByName(fieldName)
			if not field:
				raise FupCompilerError("Interface field "
					"'#%s' could not be found in the "
					"declared interface." % (
					fieldName))
			dataType = AwlDataType.makeByName(field.typeStr)
			if dataType:
				return dataType.width
		elif oper.operType == AwlOperatorTypes.SYMBOLIC:
			# Get the type of a classic symbolic operator
			# and return its bit width.
			fieldName = str(oper.offset.identChain)
			pass#TODO
			raise FupCompilerError("The symbolic operator \"%s\" is "
				"not supported, yet." % (
				fieldName))
		return 0

	def __parse(self):
		try:
			return FupCompilerFactory(compiler=self).parse(
				self.fupSource.sourceBytes)
		except FupCompilerFactory.Error as e:
			raise FupCompilerError("Failed to parse FUP source: "
				"%s" % str(e))

	def __genAwlCode(self, insns):
		"""Generate AWL code from a list of instructions.
		Returns bytes encoded as AwlSource.ENCODING.
		"""
		awl = []

		# Create header
		awl.extend(self.blockHeaderAwl)
		awl.extend(self.blockInterfAwl)

		# Create instructions body
		awl.append("BEGIN")
		fakeCpu = FupFakeCpu()
		fakeCpu.fupCompiler = self
		for insn in insns:
			insn.cpu = fakeCpu
			awl.append(insn.getStr(compact=False,
					       withSemicolon=True,
					       withComment=True))

		# Create footer
		awl.extend(self.blockFooterAwl)

		# Create the instance DBs
		awl.extend(self.instanceDBsAwl or [])

		return ('\r\n'.join(awl) + '\r\n').encode(AwlSource.ENCODING)

	def __compileBlockDecl(self):
		"""Compile block declaration.
		"""
		self.blockHeaderAwl, self.blockFooterAwl, self.instanceDBsAwl =\
			self.decl.compile(self.interf)

	def __compileInterface(self):
		"""Compile block interface.
		"""
		self.blockInterfAwl = self.interf.compile()

	def __compileGrids(self):
		"""Compile all self.grids
		"""
		# Compile the grids
		insns = []
		for grid in self.grids:
			insns.extend(grid.compile())
		insns.append(AwlInsn_BE(cpu=None))

		if self.optimizer:
			# Optimize the generated instructions
			insns = self.optimizer.optimizeInsns(
				insns,
				infoStr=str(self.fupSource))

		return insns

	def __trycompile(self, fupSource, mnemonics, optimize):
		self.reset()
		self.fupSource = fupSource
		self.mnemonics = mnemonics
		self.optimizer = AwlOptimizer() if optimize else None
		self.opTrans = AwlOpTranslator(mnemonics=mnemonics)
		self.awlSource = AwlSource(name=fupSource.name,
					   filepath=fupSource.filepath)
		if self.__parse():
			self.__compileBlockDecl()
			insns = self.__compileGrids()
			self.__compileInterface()

			# Store the AWL code in the AWL source object.
			self.awlSource.sourceBytes = self.__genAwlCode(insns)
		return self.getAwlSource()

	def compile(self, fupSource, mnemonics, optimize=True):
		"""Compile a FupSource.
		mnemonics is either MNEMONICS_EN, MNEMONICS_DE or MNEMONICS_AUTO.
		Returns an AwlSource.
		"""
		if mnemonics == S7CPUConfig.MNEMONICS_AUTO:
			try:
				return self.__trycompile(fupSource, S7CPUConfig.MNEMONICS_EN,
							 optimize=optimize)
			except AwlSimError as e:
				pass
			return self.__trycompile(fupSource, S7CPUConfig.MNEMONICS_DE,
						 optimize=optimize)
		return self.__trycompile(fupSource, mnemonics, optimize=optimize)

	def generateCallTemplate(self):
		"""Generate template AWL code for a CALL operation
		to this block.
		Returns an AwlSource.
		"""
		if not self.fupSource:
			raise FupCompilerError("FUP/FBD source is not compiled.")

		awlLines = []
		awlLines.extend(self.decl.generateCallTemplate())

		paramLines = self.interf.generateCallTemplate()
		if paramLines:
			awlLines[-1] += " ("
			awlLines.extend(paramLines)
			awlLines.append("\t)")

		awlString = "\r\n".join(awlLines)
		awlSource = AwlSource(name=("CALL " + self.fupSource.name))
		awlSource.sourceBytes = awlString.encode(AwlSource.ENCODING)
		return awlSource

	def __repr__(self):
		return "FupCompiler()"

	def __str__(self):
		return "FUP-compiler"



from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *


__all__ = [
	"VERSION_MAJOR",
	"VERSION_MINOR",
	"VERSION_STRING",
]


VERSION_MAJOR = 0
VERSION_MINOR = 57
VERSION_EXTRA = ""

VERSION_STRING = "%d.%d%s" % (VERSION_MAJOR, VERSION_MINOR, VERSION_EXTRA)



# AUTOMATICALLY GENERATED FILE
# DO NOT EDIT

# Generated from pictograms-protections.svg

icon_glasses = b'iVBORw0KGgoAAAANSUhEUgAAADAAAAAwCAYAAABXAvmHAAAA'\
               b'BHNCSVQICAgIfAhkiAAAAAlwSFlzAAAOxAAADsQBlSsOGwAA'\
               b'DPZJREFUaIG9mnl41OW1xz+/32yZmUxIJpmsZCGBbIQEEmIg'\
               b'i4oUWVSqpU9xQVEQ3IrtFb1qvdU+vVKlovQWZCmVqmAFBdpH'\
               b'qyxCCWYBQgkhYQkhIZlJyD7JJDOZZJbM7/7BIiJJZqD3fv+b'\
               b'mXPOe77vOe97zvu+I/DvQQQwQxCETCAdiFUqlSE6nU7ndrvd'\
               b'DofDMTAw0AoYgTOSJB0GvgUu3urAwi3o6oGFgiAsSEhImDR7'\
               b'9mwhZ0o+kyZlEhMdiUajpn/AgUwmolDIsdnsmBovcrLiBEXf'\
               b'HmT//v1SfX19mSRJ24EPge7/LwIRgiD8V0hIyOLHHluoWvLU'\
               b's+j1esqq2qiut1LX1ENLex+9fU48HunSIAIE6vyIMGgYGxNI'\
               b'cpw/2WnhNF1s4sPNm/jLXzbbrVbrB5IkrQDa/q8IKICXQ0JC'\
               b'Xnvl1df8Fi1eTGl5C7tLLnKmzowk+TIsiKLAhMQQZuVGkD0h'\
               b'nI0b1vP2WytsVqt1BbAKcP87CaSKovjpwoWPp6/8/SoKj7Ww'\
               b'bc8FunoGfPN6CAQHqnnknnhy0kJY/sLzbN++/ZgkSQuAmpF0'\
               b'vSHwE71e//Enn36mDR+dxOotpzG1WG/d6xtgzOgAXlo4gYry'\
               b'Ep5eurivt7d3PvDVcDojEXg6JSVl3T++2iscOtHNli/PI/mQ'\
               b'K2o/OS6XB/egx2sduUzkifuTyBirYM6sGYMmk+k5YONQ8sMR'\
               b'eC4jI2Pt13u+YcPn5yk63jzi4IE6FVnjQ4kM9Sc/M5LEuCBs'\
               b'dhf7So0cP93OyeoO+vpdXhGZmR/DgtmxTL/rdqmmpmYp8Gdf'\
               b'CPx43Lhxuw4VlYqrt57lWNXIG4MgCLz5/FRyJ0UOKePxSNSa'\
               b'LHz69TkKy5pGtFmQFcWTD8RxR0Gup6Gh4QHgC28IpAQHBx8/'\
               b'UnZcvXV3M4eOjVxrBAGeeGA8j85N4bzRwrjYwGHl7f0ufvbC'\
               b'115FY2ZeDPfmBTI1J8tqs9kmc93Cll0nrxRFcfe2z3ZGn21S'\
               b'8sXBhiENq/3k3DE5iuk5MTw5L43pU2M4dd7M878rZGKSgdBg'\
               b'DcIQ8RVEAZVCxrFTI0e2rrGHMaODuefuPNWuXTtygQ+Aqwvx'\
               b'egKvLlq0+OG75/yMVR9VDru3L543nmcfyiA9KQSDXs2Fxh5e'\
               b'ea8Y+4CbPcUNtHT0kTQmCK1agSTBjr01pMTrGfRIfPrVOR6c'\
               b'k8zpOjPN7X0jkjhRbebJh/JoNNZGVldX24DSq5NxjVxkUFDQ'\
               b'+epztZrlq8poarMNaTA4UM3WlTMpP9POnmIjTtcgp86bf5AS'\
               b'oiiQnRZG0hg9FWfbEUSBu3Nj2Fti5LWnbkOrVrBsxUHqm3pH'\
               b'JJEYF8QbT6WTlBjfZ7PZxgAdcE0EBEF4680Vb+WZHaH8s2z4'\
               b'HWf545kEBfjx0qoiak0WLrbZcLl/uFVKElxss3HyXAdtZjtt'\
               b'nXZKT7TQZrZzuKKFgswoZhfEUXisCXv/8IXXbBkgPnoUkQa1'\
               b'sri4WAbsg+8ioA8KCmqqu2BUL3q9EEuvY0hDYcEa/vrObF5+'\
               b't5h/nfapbfkBRulUrH75DtyDHpatOIjDOTisfHiIlrW/mkpc'\
               b'bLTdZrNFAL3i5d8ef3LJEnVJefOwzgMoFDIsVsctOw/QY3Xw'\
               b'5sajjI0JJCc9fET51s4+qmrMzJ8/XwMsABABBEF4cOHCxewp'\
               b'HblYdXTZCdSpiDBob9H9S2i83JaMNHFXsLukhaVP/RxBEK4S'\
               b'iE5MTMwOMYRypq5rRAMO5yB1jT1kpYbevNfXICxEiyCAqdW7'\
               b'/qr8TDspKUkYDIYcIFgECqZPn86/Trd73eecrjUzISnk5r2+'\
               b'BlGhWiy9Dq8j4B70cPxUMzNnzhSBH4mCIEzNv30a1fXed5gl'\
               b'5c3ERQbcpMvfR2iwhvqLPT7pnDPayZmSiyAImSKQkp6eQW2j'\
               b'90YuttvQqhWkJuj5yYyxTMuJRhiq7A4BjVqBTCag0yhpah26'\
               b'5twItY0WMrOyATLkQFzM6EhaO6q9UlYpZfz84QwCA1Q8fE8y'\
               b'sVE6woO1vPhEFvZ+F34qOe1mO0cqW2jttOPxSKhVcnT+StRK'\
               b'GYOShEwUyZsUgVatYHdRAxq1wicCze19xMYkA8TI1Wp1uEar'\
               b'psfmXQ6OHxtMSoKel98tJic9nNsmhLNpxynUfnKsfU4czkHU'\
               b'KjkP/CiB3UVGosK05GdGsXPfeZxuD6F6DSFBajR+cqrru8nL'\
               b'jEQhF9FplVj7nF75YLE6CNYHIgiCQa5UKpVOp9vrM22v7bvD'\
               b'+qNzU5AkCad7kGfuT78q887m4wQGqPDXyAkO9ONCYw8qpYxZ'\
               b'+XF4JImzdV2s21ZF/OgAvjh4gfTEEB65N5kN2yu98qF/wIVK'\
               b'pUChUASIMplM7sspq67RgscDXT0D2Afcl3P/+/k/SqfE3u9m'\
               b'w/YqPttzHlu/i3CDlsAAFfpRfuROikApF3G6PaQnhmAfcNPZ'\
               b'3e+1D1fWm9vtdolWq7XXT6VEJvNuEUoSFJVfJDVBT/flQ31L'\
               b'R9/VqAC43B767JfSwd7vwl+joMvy3QVAm9lOZJiWPruL88Zu'\
               b'bp8cRcmJkYvoFSgVMtzuQTwej1N0uVydtj47Oq3SawNV5zqZ'\
               b'mGzgV38ooayqlUnJBs7UdTHgcNPUZsPpGry6q5larCTGBlFr'\
               b'slzVrzX1MDHJQGZqKHWNPeg0Slo6Rm6rryAkSE1LWwdAiwiY'\
               b'jKYmokL9vTZQWdPJ5LQwTC1W3vuwnJl5sTicg+wtMVJrtPDT'\
               b'GWNZv+1SPtvsLo5WtjDn9jGs+aSC1R+VU366HbWfglfeKyYz'\
               b'JZTq+pE7gGsRHqzBZGoCMMqByooT5dPio5M4XWv2ykBXzwAO'\
               b'5yAJ0aOoa+zhwRe/Jis1lFE6FR5J4u0/H/teZ/mb948wfUo0'\
               b'd+fFEqrX0Ga2s/KDYwCkjQv2mcDY2FGcKC8BOCOXJOn44dIi'\
               b'5j6YxZcHvTdiarHym+emUFbVxppPKiitaBlW/sCRRg4cabz6'\
               b'OVCnIm1cMFnjw/iysN4nAsmxWtbvOoQkSaUi8M2ePXuk7AkR'\
               b'PlXT3647yqYdp5hze5xPg1/BOy8VsHJ5AWfqzJy94H0EFHKR'\
               b'ianhFBYWeoASEWhtaGgob21tZUJisNeG3IMeio434x6UMOjV'\
               b'PhMwBKnZvOs0r64uYcDh1TUoAJPTwqioqKS9vb0EaBMBJEna'\
               b'8qcNa5mTH+WTE5IkUdPQzcRkg096CrlIgL+KouO+Pw/Myo3g'\
               b'w81/4vK1PFdOZB9v2bLFcduEcJ9n8+yFLrLTRj5NXYvIUH+s'\
               b'fU46uuw+6Y0O8ycxNoBt27b1AlvgOwLdFotl0/p17/Pw7ASf'\
               b'jB471cbEZMOQd0A3QnJ8EPVNPT5fyT96bwLvvfsODodjE9AL'\
               b'3xFAkqQVq1b9vn9qhoGE6FFeG60810lXzwA/nZnolbxOq2Tu'\
               b'tARO1nT65Hxqgp5xMWrWrPmjRZKkt698L14j09rd3f3f//HL'\
               b'ZSxfmOZDayHxxtrDzMqPoSBr+DUkCPD6szmEBav5+/46r51X'\
               b'KWUsX5jGc88sZWBg4E3gKvvrvZSLolj2+Y6/T3Iq49n4+Vmv'\
               b'BwnwV7Lu13dx4IiJLV9W477unigwQMVrS28jMEDF62sO+9Q6'\
               b'/OKRNFovFPPsM0sPS5JUAFytkjea5uSAgICysmMndDsLO/im'\
               b'tPEGIjeGQa/mPxdNZnS4P5/vPU9VTSd+KjkZSQbmzRjL7qIG'\
               b'Nv/t9A/IDYf7p48hb7yCOwqm9trt9kzge6Eb8no9Pj5+V+G3'\
               b'JeKmXRe8ehu4FuNiA5mSEUFspA6NnwKr3ck/jzRytLLVJzsz'\
               b'cqN5aGYUuVOyB1tbW+8B9l4vM1yiL0lKStq4/8Ah4ZPdRvYU'\
               b'm3wa/FYgCPDju8ZwX34od00rkIxGo88PHFfwdGxs7Nqvdu+T'\
               b'VV1w8cHfzvkU/puBUiFj2UOpRAQ5mXvvrMHm5ubFwEdDyYtD'\
               b'/XAZG4xG4/25U7L73JYq1r6ay9iY4R8vbgUTEkPY8Ho+7caj'\
               b'FOTlWJqbm+9jGOfB+2fWREEQts6bNy979R/WUlFjYes/6mgz'\
               b'+1ZJh8LocB2P3ZdAUoyWp5YuYu/evUclSXqE6xbsjeDLZY4c'\
               b'eEGr1f76heUv+i9b9gsqa8zsO9xK+dl2n1NLIRfJnhDO7LwI'\
               b'kuICWfn271i37n1rf3//G8D/AF4ZvJm/GoQKgvCqWq1evGDB'\
               b'At3CJ5aQmpLMiTOtVBv7qDX10Ga209ndj8vtQZIkNH4KAgNU'\
               b'jA7zJz56FClxWialRnCy8hQb169h586dff39/eskSVoJeHeq'\
               b'ugUCVzAKeEwQhPmRkZG5d955p5CffwcZk7KIiR5NWFgwctml'\
               b'95MBhxOz2UJ9QwPlx8s4criUffv2ebq6uo5IkvQx8Ffgpl7P'\
               b'b4XAtQgH8gVBmAqkAHFAuCiKKrlcLnc6nT1cehIyASclSSoH'\
               b'DuDjbN8I/wuzJT0nIDcjfQAAAABJRU5ErkJggg=='



# AUTOMATICALLY GENERATED FILE
# DO NOT EDIT

# Generated from appointment-soon.svg

icon_timer = b'iVBORw0KGgoAAAANSUhEUgAAADAAAAAwCAYAAABXAvmHAAAA'\
             b'BHNCSVQICAgIfAhkiAAAAAlwSFlzAAAOxAAADsQBlSsOGwAA'\
             b'EFZJREFUaIHdmXmQXEd9xz/9jrln7x2t9tC50q5kGR1GRgLL'\
             b'NsKyiQHHZSSgMOJyiAMh4CQmRRIbCFdxmsJOChISymBkDDFQ'\
             b'IQRjrNjGlg8c2ZIsydJqJa2klfY+Z+ee1/3LH/Nmd/awjQMk'\
             b'VXlVv+qenvf6fT+/36/7db8Hv8dDPon1++wfQP0uOjmyi1gh'\
             b'xBss1BXKcV4nYpYabWoQCaKUVkrllaUmsOimaJ42yN5clide'\
             b'ez/Z/zMA+STWc6e42nacW7WRy2qWrMjH29fFw4lWFahKYIWj'\
             b'WE4QEUG8PDqbojg5RHrgnJk8eTiV6e91LdQD2nh3bryHxxTI'\
             b'/wqAgHr2Rq53XPtroYamuqatV8VjKzahTAFJX4DMCFKYBC8D'\
             b'xitdZDngRFHBaogkUNFmjMDkiWdl8OmHU97k8Dmt9S2b7mHv'\
             b'7xXgmXeyIhCw94Qbmi5qvfpt8VBiOTJ+DBk/AflkCa/sx7n+'\
             b'VJWlglAtqnY1qqaDTN9xen95f6owMfp4sejdtHkP/b9zgAPv'\
             b'5m3Kdr/Vds3OWE3nVkuGDyCjx0C0L1goZAqkx1OkkzkKOYOn'\
             b'DbalsBxLQiFLhWMhovVx3KADSpUo7SCq/iKoX8fYoYf1hUd/'\
             b'lqTgvWPjHn75OwEQUAffa38+WNPw4ZU7PxSzVRbpexK8LGDh'\
             b'FTQj54Zl9MI42kMVFq0lX9+BxBohEENMASufwZrqJzB6Anfo'\
             b'OG40JA3NNdQvaVCWbQEG3Ciq9XKKuSLd//qNlJ5KfmLjd83X'\
             b'fisAAfX8++xvhJva3rVi50eijB4qeV0pjIaBk/0ycm6Y3NLL'\
             b'VXbNW9B1K7CUheM62LaNskrdixa00XjFIoLBHjhK+IWfEu5/'\
             b'ThqXJUi0L1YKAwgqsR5TtZpT992Ryo8OfX3D3fq2/zHAwffZ'\
             b'n4k1td2ybNctMfoeQ6bOg+WSHk3Sc7BHMm2XqfSm92JHawlH'\
             b'IoSsIvbJX2FfeA7Gz2EmB1A1rTi770YAYwyFQoFcLkc2k0En'\
             b'+4n/+p+IjB6VlZesUMF4GKSIqlkJiUvpvvfLqfzI4N9u+I65'\
             b'8xUDHHg3bwvWJf5l1e6Px1T/PiR1HqwQg6cGZODMMBNX3K5U'\
             b'yzqqqqoInt2H2n8P+sIxGla24IYsArEwbjhI994DRD91FEFK'\
             b'U6pvxmgymQzJZBJ19mlq931VWte0UtvaqNA5VM1KTMNGjn/7'\
             b'86nC1MQfvvp7PPwbAxx8F8tVIHCw86bbqux0NzLaBU6E/q5e'\
             b'GRjSTLzxCyrW0EI8EoR730+McarihnhTHcoNQiCAcgOICEd+'\
             b'8gRVnzkOUBIugojBGIOIQWtNMpkk1d9D3S8+Js3L60isWKzw'\
             b'MqjEeopWA8fv/vKY0t6aDd9jaK7WeY96ASWuvafl6l1xW9LI'\
             b'6HFwogyfHpSBQY/km+5Q9S3LqUqdofjFLTRZ51icEGLVIcTz'\
             b'EO2B1mD0dJ+2XRoT02aVSsuysW2H6uoa6pZ2Mn7dXaqvZ4zx'\
             b'8yOCE0GGDuGGXVqvfHOVcp1/XMjZ8wAO7Oa6SH3zutrOS5X0'\
             b'PwFOiMx4ir7TAySv/ZKqb24jdHYfzv0fpLM9SCxqgechni9c'\
             b'e6A9xJsBcGx7BsKqgPDNsizC4RD1rSuYvPYr6vwLZ8inC2AF'\
             b'kAuPU7txu+PEaq7afyPbXhJAQFm2fUfrNW+Py9CzoIsIUXr2'\
             b'd8nklbepupbluH3Pwc9uo6065Qv1kErhWvswPlA5ApaNbTnT'\
             b'nrds3yzLN5uAG6BmaSfjW/5c9TxzXHBiUEzD6BHart4ZcwP2'\
             b'V18S4OCN7IgkWhtDDa3IWBe4VQyc6JFU6xYVWrmZ/b96CP3D'\
             b'j7KsvrQGE2+OcO2BHwnRfjrBguljTwv3TSksSxFwXaIbrmEq'\
             b'voyR0+cFJ46MHiW65CKceP2aZ3fz6hcFsALORxJbdsRl7Cgo'\
             b'Cy0uI6d7kW0fIhaN4oVq2H3oVTxwocEPmfHF+x73vPm/p8fA'\
             b'zDgoe9xSZfEWyrJQfhkOhSlu+6ga6DqjxAqD8ZDxEzReuj1q'\
             b'2/aHFgQ4souYoLZHl78KGe8GN8Zoz3nJLr1MRRuaUUpxxeXb'\
             b'+Itb/pSvdK/kY4fWlhi0Lgn1KqKgPfD0dAQce2bQWpZVyntf'\
             b'vKqAsJTCUgqlFLG2TrL1HST7+sGNIuMnqOnYrAzs+uEu7HkA'\
             b'BZdt0SWrPMtLQTGNcuOMnu1TbNxZ8pDf8datm/nAB/+Yw+kG'\
             b'tj9yKYfHorOF+2NAFhwDJc/PEm2p6b6Vfx+UwrJtzPobGDo7'\
             b'KMqthtwYtqMI1TdLe4BN8wCUo66qbl8XlXQf2C7FnEcxl5dg'\
             b'28UwfQNQKK68fBu7372beHUNf3V0HV860jaT9xWzkfgAZc/P'\
             b'pE2FaGbq5ZVqqQ2CHdvIDQ8pLTYoC0ldoK7j4pDYbJ8HYDnO'\
             b'peFEm0VmCOwomdERvMXrlGXZ/tNOlUwBSnHNNW/ghhtuIFLd'\
             b'wK/z7Vz38Dr6knbFoK4YA1Y5XdQ8K61Ky3fwneSDWG4Qr26Z'\
             b'ZMfHwIlCdphwotV1bOfSeQAipj0Qq0Pyk+CEyCbTmMZO5h4l'\
             b'/5SO6667lmvf9AcoJ0i0oYXdj3Xw3a5aX7w/Q8F0alR6eaa/'\
             b'cmX+fwC6YZXKJ1NgB5HcBG51PcDa+QBaqq1gCLwMygqQy+Sx'\
             b'61rnAcicncquG65nx44dFLTQuqSVH/W28Z6H2sjkNQNJmb5G'\
             b'pLSUKPfC3JrIwpvK2qXk0jlRdgC8NHYgghGpK//tTF9vTFDZ'\
             b'DmKKYLmIZ1DBWIXgipvKjCBBuPGdb8fzPB599FGqGxqZSkW5'\
             b'+t/C1IbA1fV85fQpli5ZUrGYY7pe2ffMgm+mzQrH0CMGlAvG'\
             b'Q9kWGInMA0ApQbxy3Y+kzBcsJaTKlaWI8L737KZQLPDUk08x'\
             b'ODZB2+Im6uuqGRwa5r1/dDPvufGdrFrVTnPLYhYlGuddPxum'\
             b'ImICWP7AECk9eyq86czoV3lTyEaUVWpSbhDJp6e9PHs5bBCx'\
             b'Zn4bg0HxJx+4ia6ubpJTKS7ZuJ7GxgQPP/KfxCIF9tx3H6FQ'\
             b'iFyuQCaboaG+gZaWxSxbtoTWlmYWLUqweHGCSCTsCy+BkJ/C'\
             b'Dbglh1oOUsihLJWZD2CpCZOdith2EMQjHAsjk30LizaCKIOI'\
             b'wpiKYa1g86ZLON97jmPHj3Hy1EkAbNvCdWxi4SAh1yEWCZAv'\
             b'5Dh9+jRdXScoao1X9OjsWMXtt906KyrWRC+hqphCF8AO4mWT'\
             b'KNTo/BSy6C5MDDWH3Sji5YjU1uL0HEVPe7lCtCiUKJQpTYGV'\
             b'88Zbb7iO7+65l3g0QzAYoFj0GBtL4rguoYjQ0JhgUaKRpqZF'\
             b'NNTVUVdfS21tDbW11di2jTZ6es8gItjDJwguvQjRWVQwTmFi'\
             b'CCyOzAcomqfTQ/3bIssbLUn2El7UjnpiH2I0Ys0WrVAYpWZJ'\
             b'9zMXy1L8w11f5W9u/zQFXWBqKs1nPn07Gzesx4hBjGDEYIzG'\
             b'GFMSrEulNhrxNzpGBOPlcUfPEG7cAWMHILaYzGB3wSt6v54H'\
             b'YJC9yTNdH2zs7Kii2IUTiRKIxqXYf0SZlvUoJRgjKExpPjdz'\
             b'N3N+pJSipXkx3/n2Nzl2vIu1nZ0Igqe9iu2kmQVhpFTKdHsJ'\
             b'Qp16kkhzmyglSnQRQnWMn+zOW3pmezkNkMvyhOrrdQ0uyglA'\
             b'dpjGzk6yB3+Mbr7YTxf/KTlHfFmYUlKKjG+rV7Xj+Q+zyrFk'\
             b'ygPfF69NBUw5AsbgPP8jEqs7FNkhcKN4+QLe5LBsWMXzM5nv'\
             b'H6+9n6xl8cBkz3FRkUVIuo/a1euVc+ZJkeTQdOfGv5n2w6+1'\
             b'rig9v+4tbMZDa40pX2M0WhuM1nP6N8jIKdzxHuLL1iCZflR0'\
             b'ERMnj2il1B71d5h5AADa8+4cOrg/TbQZvCyW5Fj0qo24j/99'\
             b'hYf80hdi5kLokkhPl39rPL9t+jwfRBs9ayzoir7dR+6Q5k2X'\
             b'CoVREI2EGhk68Fy2UNTfqNQ8C2DjPTzmTQ6fzQwPQKQRmegi'\
             b'seG1KjB4SDj7X75YUyGkUnylyJLHPd8qwaZLMzsSldGV4w8R'\
             b'KY5Su2aTkoluiDYxebZHpJh6dvO9HH5RAAWiPX3L+X2/ShFt'\
             b'BZ2H9AVW7niLCj34aZGpId9jC3l+JnU8PQdIe3hzIbU3+9oy'\
             b'yMgZwo99XVbsuE4xcao0fkIJ+p96LO0V9a1zZo75byU23cPe'\
             b'wuTY4yMnjmkVb0MmuglW17LkNVsJ3v9nojPj0zc2c9JmJnW8'\
             b'WUDTEAullN+P0RqZGiT8k1tkxet3KMdVyNQZVHwJI4cPFEw+'\
             b'8+NL7mH/ywIAFIveTQPPPJnM5xWEa5GhZ6ju2KRaL+og/IOb'\
             b'xUz0L+C9crp40yK9BdLIe5EomJEewvfdLEu2bCHa1o6MHIBI'\
             b'gmwyy+ChA+OqoG9ZSKu9UOO3DpN6f6c5NHWh9/r6tZcElE7D'\
             b'1Dkiqy5X4ZBN9hd3Ka9mKaa6ddZSeWa9ZBZYrBmMqXgG+FOl'\
             b'0Rqr6yEiD35KVm7foaqWr1XSvw8CETy7ju7/+Hkqncu/efP3'\
             b'6f6NAQD3n4/Q+472YjE/cGFL3ZqNAVWcguQpQks2qboVHaQf'\
             b'uVtU73OqmOhAArHZ7z2nIWZseu6XmamS0R4iD36K+IWnWH39'\
             b'jSpUXYP07wM3jA4kOPnzn6dGJzK3XvEDeZDScs7MFboQgEvp'\
             b'AefefUQO7lqRj+X7zl9cs/rigKU8GOvCjtXRuPlNKpAdkMLe'\
             b'byq77zCeE8XEEgjKF+wvB6S0dChHQLwc1uknCT1+F+ED36dt'\
             b'w3pp3f5WZaXPltImVIOnqjj5wC/S/WOpL1/1Q/OdCm3zIOau'\
             b'BywfYBoCcP/9rermpY3Rv1zx+iujwTCQHQEngmrchATqSJ3c'\
             b'z8Dhg+QGzyONKynWt6NrWiAQBRFUPo09eR57pBtr9Azh1mU0'\
             b'rd1AbOUmyPYjwwfAFCDcSHYqz6lHHk+fGsp/YudPzX1A0Tev'\
             b'om5eEQDg3vl6rrhsiXNn04Z10cZVKx0KE1BIQSCOqlmFqm7H'\
             b'FPPkhnrIjw6SS06ii0UAnGCAULyWYP0iQollKMtCJk8ik91Q'\
             b'zEAgjrhVDB/vKgwcOT71y9P65r9+nP0Vgn9jgFkpNLe+cyVN'\
             b'H36N89m6eGhL6+YNkfjiRshPQTEFxkCwGhVeBMFqcKNgBQAB'\
             b'Uyydk59EMv0lcMsuRciJMt47KAPPPZ8dTeUe+twT5rN7exl9'\
             b'EeHl+oum0EIQlaUDOF+6jK3bltkfj0bDyxs620O1S5otx7VA'\
             b'50AXS59XRfu7Kvwtqg22W/rkaoco5jXjZ8/rkWOn8ul8/uhD'\
             b'p/UXP/kkB32RZSsuUM46XuoTk10pusLK7fbnXsfGrW3WjTUB'\
             b'9UYnGjM1LYlgpK7aDcSjOIEglmOX9glFj2IhT2EqTWZ0vDBx'\
             b'YbjgZdJM5tVP957R937+GV4AtC+yXM41PVfgywFUnuNWCq8w'\
             b'C7AXBQl87DVcvLqe19VFnPUhm1VKSbUSCQmglMpqURN5Q9do'\
             b'2jt4bIgnvvAsL0zkpoUZvyxbub3Iy3zBf6Vf6pUv2imL90ur'\
             b'4r+ZV20z5exXDvifJEtlWbzxhRteRvRvA/BS/VSKr4QoHzLH'\
             b'XpHQ/7fHfwMs8lEpPuwWGAAAAABJRU5ErkJggg=='



#!/usr/bin/env python

import math

# Python code which draws the PuTTY icon components at a range of
# sizes.

# TODO
# ----
#
#  - use of alpha blending
#     + try for variable-transparency borders
#
#  - can we integrate the Mac icons into all this? Do we want to?

def pixel(x, y, colour, canvas):
    canvas[(int(x),int(y))] = colour

def overlay(src, x, y, dst):
    x = int(x)
    y = int(y)
    for (sx, sy), colour in src.items():
        dst[sx+x, sy+y] = blend(colour, dst.get((sx+x, sy+y), cT))

def finalise(canvas):
    for k in canvas.keys():
        canvas[k] = finalisepix(canvas[k])

def bbox(canvas):
    minx, miny, maxx, maxy = None, None, None, None
    for (x, y) in canvas.keys():
        if minx == None:
            minx, miny, maxx, maxy = x, y, x+1, y+1
        else:
            minx = min(minx, x)
            miny = min(miny, y)
            maxx = max(maxx, x+1)
            maxy = max(maxy, y+1)
    return (minx, miny, maxx, maxy)

def topy(canvas):
    miny = {}
    for (x, y) in canvas.keys():
        miny[x] = min(miny.get(x, y), y)
    return miny

def render(canvas, minx, miny, maxx, maxy):
    w = maxx - minx
    h = maxy - miny
    ret = []
    for y in range(h):
        ret.append([outpix(cT)] * w)
    for (x, y), colour in canvas.items():
        if x >= minx and x < maxx and y >= miny and y < maxy:
            ret[y-miny][x-minx] = outpix(colour)
    return ret

# Code to actually draw pieces of icon. These don't generally worry
# about positioning within a canvas; they just draw at a standard
# location, return some useful coordinates, and leave composition
# to other pieces of code.

sqrthash = {}
def memoisedsqrt(x):
    if not sqrthash.has_key(x):
        sqrthash[x] = math.sqrt(x)
    return sqrthash[x]

BR, TR, BL, TL = range(4) # enumeration of quadrants for border()

def border(canvas, thickness, squarecorners, out={}):
    # I haven't yet worked out exactly how to do borders in a
    # properly alpha-blended fashion.
    #
    # When you have two shades of dark available (half-dark H and
    # full-dark F), the right sequence of circular border sections
    # around a pixel x starts off with these two layouts:
    #
    #   H    F
    #  HxH  FxF
    #   H    F
    #
    # Where it goes after that I'm not entirely sure, but I'm
    # absolutely sure those are the right places to start. However,
    # every automated algorithm I've tried has always started off
    # with the two layouts
    #
    #   H   HHH
    #  HxH  HxH
    #   H   HHH
    #
    # which looks much worse. This is true whether you do
    # pixel-centre sampling (define an inner circle and an outer
    # circle with radii differing by 1, set any pixel whose centre
    # is inside the inner circle to F, any pixel whose centre is
    # outside the outer one to nothing, interpolate between the two
    # and round sensibly), _or_ whether you plot a notional circle
    # of a given radius and measure the actual _proportion_ of each
    # pixel square taken up by it.
    #
    # It's not clear what I should be doing to prevent this. One
    # option is to attempt error-diffusion: Ian Jackson proved on
    # paper that if you round each pixel's ideal value to the
    # nearest of the available output values, then measure the
    # error at each pixel, propagate that error outwards into the
    # original values of the surrounding pixels, and re-round
    # everything, you do get the correct second stage. However, I
    # haven't tried it at a proper range of radii.
    #
    # Another option is that the automated mechanisms described
    # above would be entirely adequate if it weren't for the fact
    # that the human visual centres are adapted to detect
    # horizontal and vertical lines in particular, so the only
    # place you have to behave a bit differently is at the ends of
    # the top and bottom row of pixels in the circle, and the top
    # and bottom of the extreme columns.
    #
    # For the moment, what I have below is a very simple mechanism
    # which always uses only one alpha level for any given border
    # thickness, and which seems to work well enough for Windows
    # 16-colour icons. Everything else will have to wait.

    thickness = memoisedsqrt(thickness)

    if thickness < 0.9:
        darkness = 0.5
    else:
        darkness = 1
    if thickness < 1: thickness = 1
    thickness = round(thickness - 0.5) + 0.3

    out["borderthickness"] = thickness

    dmax = int(round(thickness))
    if dmax < thickness: dmax = dmax + 1

    cquadrant = [[0] * (dmax+1) for x in range(dmax+1)]
    squadrant = [[0] * (dmax+1) for x in range(dmax+1)]

    for x in range(dmax+1):
        for y in range(dmax+1):
            if max(x, y) < thickness:
                squadrant[x][y] = darkness
            if memoisedsqrt(x*x+y*y) < thickness:
                cquadrant[x][y] = darkness

    bvalues = {}
    for (x, y), colour in canvas.items():
        for dx in range(-dmax, dmax+1):
            for dy in range(-dmax, dmax+1):
                quadrant = 2 * (dx < 0) + (dy < 0)
                if (x, y, quadrant) in squarecorners:
                    bval = squadrant[abs(dx)][abs(dy)]
                else:
                    bval = cquadrant[abs(dx)][abs(dy)]
                if bvalues.get((x+dx,y+dy),0) < bval:
                    bvalues[(x+dx,y+dy)] = bval

    for (x, y), value in bvalues.items():
        if not canvas.has_key((x,y)):
            canvas[(x,y)] = dark(value)

def sysbox(size, out={}):
    canvas = {}

    # The system box of the computer.

    height = int(round(3.6*size))
    width = int(round(16.51*size))
    depth = int(round(2*size))
    highlight = int(round(1*size))
    bothighlight = int(round(1*size))

    out["sysboxheight"] = height

    floppystart = int(round(19*size)) # measured in half-pixels
    floppyend = int(round(29*size)) # measured in half-pixels
    floppybottom = height - bothighlight
    floppyrheight = 0.7 * size
    floppyheight = int(round(floppyrheight))
    if floppyheight < 1:
        floppyheight = 1
    floppytop = floppybottom - floppyheight

    # The front panel is rectangular.
    for x in range(width):
        for y in range(height):
            grey = 3
            if x < highlight or y < highlight:
                grey = grey + 1
            if x >= width-highlight or y >= height-bothighlight:
                grey = grey - 1
            if y < highlight and x >= width-highlight:
                v = (highlight-1-y) - (x-(width-highlight))
                if v < 0:
                    grey = grey - 1
                elif v > 0:
                    grey = grey + 1
            if y >= floppytop and y < floppybottom and \
            2*x+2 > floppystart and 2*x < floppyend:
                if 2*x >= floppystart and 2*x+2 <= floppyend and \
                floppyrheight >= 0.7:
                    grey = 0
                else:
                    grey = 2
            pixel(x, y, greypix(grey/4.0), canvas)

    # The side panel is a parallelogram.
    for x in range(depth):
        for y in range(height):
            pixel(x+width, y-(x+1), greypix(0.5), canvas)

    # The top panel is another parallelogram.
    for x in range(width-1):
        for y in range(depth):
            grey = 3
            if x >= width-1 - highlight:
                grey = grey + 1         
            pixel(x+(y+1), -(y+1), greypix(grey/4.0), canvas)

    # And draw a border.
    border(canvas, size, [], out)

    return canvas

def monitor(size):
    canvas = {}

    # The computer's monitor.

    height = int(round(9.55*size))
    width = int(round(11.49*size))
    surround = int(round(1*size))
    botsurround = int(round(2*size))
    sheight = height - surround - botsurround
    swidth = width - 2*surround
    depth = int(round(2*size))
    highlight = int(round(math.sqrt(size)))
    shadow = int(round(0.55*size))

    # The front panel is rectangular.
    for x in range(width):
        for y in range(height):
            if x >= surround and y >= surround and \
            x < surround+swidth and y < surround+sheight:
                # Screen.
                sx = (float(x-surround) - swidth/3) / swidth
                sy = (float(y-surround) - sheight/3) / sheight
                shighlight = 1.0 - (sx*sx+sy*sy)*0.27
                pix = bluepix(shighlight)
                if x < surround+shadow or y < surround+shadow:
                    pix = blend(cD, pix) # sharp-edged shadow on top and left
            else:
                # Complicated double bevel on the screen surround.

                # First, the outer bevel. We compute the distance
                # from this pixel to each edge of the front
                # rectangle.
                list = [
                (x, +1),
                (y, +1),
                (width-1-x, -1),
                (height-1-y, -1)
                ]
                # Now sort the list to find the distance to the
                # _nearest_ edge, or the two joint nearest.
                list.sort()
                # If there's one nearest edge, that determines our
                # bevel colour. If there are two joint nearest, our
                # bevel colour is their shared one if they agree,
                # and neutral otherwise.
                outerbevel = 0
                if list[0][0] < list[1][0] or list[0][1] == list[1][1]:
                    if list[0][0] < highlight:
                        outerbevel = list[0][1]

                # Now, the inner bevel. We compute the distance
                # from this pixel to each edge of the screen
                # itself.
                list = [
                (surround-1-x, -1),
                (surround-1-y, -1),
                (x-(surround+swidth), +1),
                (y-(surround+sheight), +1)
                ]
                # Now we sort to find the _maximum_ distance, which
                # conveniently ignores any less than zero.
                list.sort()
                # And now the strategy is pretty much the same as
                # above, only we're working from the opposite end
                # of the list.
                innerbevel = 0
                if list[-1][0] > list[-2][0] or list[-1][1] == list[-2][1]:
                    if list[-1][0] >= 0 and list[-1][0] < highlight:
                        innerbevel = list[-1][1]

                # Now we know the adjustment we want to make to the
                # pixel's overall grey shade due to the outer
                # bevel, and due to the inner one. We break a tie
                # in favour of a light outer bevel, but otherwise
                # add.
                grey = 3
                if outerbevel > 0 or outerbevel == innerbevel:
                    innerbevel = 0
                grey = grey + outerbevel + innerbevel

                pix = greypix(grey / 4.0)

            pixel(x, y, pix, canvas)

    # The side panel is a parallelogram.
    for x in range(depth):
        for y in range(height):
            pixel(x+width, y-x, greypix(0.5), canvas)

    # The top panel is another parallelogram.
    for x in range(width):
        for y in range(depth-1):
            pixel(x+(y+1), -(y+1), greypix(0.75), canvas)

    # And draw a border.
    border(canvas, size, [(0,int(height-1),BL)])

    return canvas

def computer(size):
    # Monitor plus sysbox.
    out = {}
    m = monitor(size)
    s = sysbox(size, out)
    x = int(round((2+size/(size+1))*size))
    y = int(out["sysboxheight"] + out["borderthickness"])
    mb = bbox(m)
    sb = bbox(s)
    xoff = sb[0] - mb[0] + x
    yoff = sb[3] - mb[3] - y
    overlay(m, xoff, yoff, s)
    return s

def lightning(size):
    canvas = {}

    # The lightning bolt motif.

    # We always want this to be an even number of pixels in height,
    # and an odd number in width.
    width = round(7*size) * 2 - 1
    height = round(8*size) * 2

    # The outer edge of each side of the bolt goes to this point.
    outery = round(8.4*size)
    outerx = round(11*size)

    # And the inner edge goes to this point.
    innery = height - 1 - outery
    innerx = round(7*size)

    for y in range(int(height)):
        list = []
        if y <= outery:
            list.append(width-1-int(outerx * float(y) / outery + 0.3))
        if y <= innery:
            list.append(width-1-int(innerx * float(y) / innery + 0.3))
        y0 = height-1-y
        if y0 <= outery:
            list.append(int(outerx * float(y0) / outery + 0.3))
        if y0 <= innery:
            list.append(int(innerx * float(y0) / innery + 0.3))
        list.sort()
        for x in range(int(list[0]), int(list[-1]+1)):
            pixel(x, y, cY, canvas)

    # And draw a border.
    border(canvas, size, [(int(width-1),0,TR), (0,int(height-1),BL)])

    return canvas

def document(size):
    canvas = {}

    # The document used in the PSCP/PSFTP icon.

    width = round(13*size)
    height = round(16*size)

    lineht = round(1*size)
    if lineht < 1: lineht = 1
    linespc = round(0.7*size)
    if linespc < 1: linespc = 1
    nlines = int((height-linespc)/(lineht+linespc))
    height = nlines*(lineht+linespc)+linespc # round this so it fits better

    # Start by drawing a big white rectangle.
    for y in range(int(height)):
        for x in range(int(width)):
            pixel(x, y, cW, canvas)

    # Now draw lines of text.
    for line in range(nlines):
        # Decide where this line of text begins.
        if line == 0:
            start = round(4*size)
        elif line < 5*nlines/7:
            start = round((line - (nlines/7)) * size)
        else:
            start = round(1*size)
        if start < round(1*size):
            start = round(1*size)
        # Decide where it ends.
        endpoints = [10, 8, 11, 6, 5, 7, 5]
        ey = line * 6.0 / (nlines-1)
        eyf = math.floor(ey)
        eyc = math.ceil(ey)
        exf = endpoints[int(eyf)]
        exc = endpoints[int(eyc)]
        if eyf == eyc:
            end = exf
        else:
            end = exf * (eyc-ey) + exc * (ey-eyf)
        end = round(end * size)

        liney = height - (lineht+linespc) * (line+1)
        for x in range(int(start), int(end)):
            for y in range(int(lineht)):
                pixel(x, y+liney, cK, canvas)

    # And draw a border.
    border(canvas, size, \
    [(0,0,TL),(int(width-1),0,TR),(0,int(height-1),BL), \
    (int(width-1),int(height-1),BR)])

    return canvas

def hat(size):
    canvas = {}

    # The secret-agent hat in the Pageant icon.

    topa = [6]*9+[5,3,1,0,0,1,2,2,1,1,1,9,9,10,10,11,11,12,12]
    topa = [round(x*size) for x in topa]
    botl = round(topa[0]+2.4*math.sqrt(size))
    botr = round(topa[-1]+2.4*math.sqrt(size))
    width = round(len(topa)*size)

    # Line equations for the top and bottom of the hat brim, in the
    # form y=mx+c. c, of course, needs scaling by size, but m is
    # independent of size.
    brimm = 1.0 / 3.75
    brimtopc = round(4*size/3)
    brimbotc = round(10*size/3)

    for x in range(int(width)):
        xs = float(x) * (len(topa)-1) / (width-1)
        xf = math.floor(xs)
        xc = math.ceil(xs)
        topf = topa[int(xf)]
        topc = topa[int(xc)]
        if xf == xc:
            top = topf
        else:
            top = topf * (xc-xs) + topc * (xs-xf)
        top = math.floor(top)
        bot = round(botl + (botr-botl) * x/(width-1))

        for y in range(int(top), int(bot)):
            pixel(x, y, cK, canvas)

    # Now draw the brim.
    for x in range(int(width)):
        brimtop = brimtopc + brimm * x
        brimbot = brimbotc + brimm * x
        for y in range(int(math.floor(brimtop)), int(math.ceil(brimbot))):
            tophere = max(min(brimtop - y, 1), 0)
            bothere = max(min(brimbot - y, 1), 0)
            grey = bothere - tophere
            # Only draw brim pixels over pixels which are (a) part
            # of the main hat, and (b) not right on its edge.
            if canvas.has_key((x,y)) and \
            canvas.has_key((x,y-1)) and \
            canvas.has_key((x,y+1)) and \
            canvas.has_key((x-1,y)) and \
            canvas.has_key((x+1,y)):
                pixel(x, y, greypix(grey), canvas)

    return canvas

def key(size):
    canvas = {}

    # The key in the PuTTYgen icon.

    keyheadw = round(9.5*size)
    keyheadh = round(12*size)
    keyholed = round(4*size)
    keyholeoff = round(2*size)
    # Ensure keyheadh and keyshafth have the same parity.
    keyshafth = round((2*size - (int(keyheadh)&1)) / 2) * 2 + (int(keyheadh)&1)
    keyshaftw = round(18.5*size)
    keyhead = [round(x*size) for x in [12,11,8,10,9,8,11,12]]

    squarepix = []

    # Ellipse for the key head, minus an off-centre circular hole.
    for y in range(int(keyheadh)):
        dy = (y-(keyheadh-1)/2.0) / (keyheadh/2.0)
        dyh = (y-(keyheadh-1)/2.0) / (keyholed/2.0)
        for x in range(int(keyheadw)):
            dx = (x-(keyheadw-1)/2.0) / (keyheadw/2.0)
            dxh = (x-(keyheadw-1)/2.0-keyholeoff) / (keyholed/2.0)
            if dy*dy+dx*dx <= 1 and dyh*dyh+dxh*dxh > 1:
                pixel(x + keyshaftw, y, cy, canvas)

    # Rectangle for the key shaft, extended at the bottom for the
    # key head detail.
    for x in range(int(keyshaftw)):
        top = round((keyheadh - keyshafth) / 2)
        bot = round((keyheadh + keyshafth) / 2)
        xs = float(x) * (len(keyhead)-1) / round((len(keyhead)-1)*size)
        xf = math.floor(xs)
        xc = math.ceil(xs)
        in_head = 0
        if xc < len(keyhead):
            in_head = 1
            yf = keyhead[int(xf)]
            yc = keyhead[int(xc)]
            if xf == xc:
                bot = yf
            else:
                bot = yf * (xc-xs) + yc * (xs-xf)
        for y in range(int(top),int(bot)):
            pixel(x, y, cy, canvas)
            if in_head:
                last = (x, y)
        if x == 0:
            squarepix.append((x, int(top), TL))
        if x == 0:
            squarepix.append(last + (BL,))
        if last != None and not in_head:
            squarepix.append(last + (BR,))
            last = None

    # And draw a border.
    border(canvas, size, squarepix)

    return canvas

def linedist(x1,y1, x2,y2, x,y):
    # Compute the distance from the point x,y to the line segment
    # joining x1,y1 to x2,y2. Returns the distance vector, measured
    # with x,y at the origin.

    vectors = []

    # Special case: if x1,y1 and x2,y2 are the same point, we
    # don't attempt to extrapolate it into a line at all.
    if x1 != x2 or y1 != y2:
        # First, find the nearest point to x,y on the infinite
        # projection of the line segment. So we construct a vector
        # n perpendicular to that segment...
        nx = y2-y1
        ny = x1-x2
        # ... compute the dot product of (x1,y1)-(x,y) with that
        # vector...
        nd = (x1-x)*nx + (y1-y)*ny
        # ... multiply by the vector we first thought of...
        ndx = nd * nx
        ndy = nd * ny
        # ... and divide twice by the length of n.
        ndx = ndx / (nx*nx+ny*ny)
        ndy = ndy / (nx*nx+ny*ny)
        # That gives us a displacement vector from x,y to the
        # nearest point. See if it's within the range of the line
        # segment.
        cx = x + ndx
        cy = y + ndy
        if cx >= min(x1,x2) and cx <= max(x1,x2) and \
        cy >= min(y1,y2) and cy <= max(y1,y2):
            vectors.append((ndx,ndy))

    # Now we have up to three candidate result vectors: (ndx,ndy)
    # as computed just above, and the two vectors to the ends of
    # the line segment, (x1-x,y1-y) and (x2-x,y2-y). Pick the
    # shortest.
    vectors = vectors + [(x1-x,y1-y), (x2-x,y2-y)]
    bestlen, best = None, None
    for v in vectors:
        vlen = v[0]*v[0]+v[1]*v[1]
        if bestlen == None or bestlen > vlen:
            bestlen = vlen
            best = v
    return best

def spanner(size):
    canvas = {}

    # The spanner in the config box icon.

    headcentre = 0.5 + round(4*size)
    headradius = headcentre + 0.1
    headhighlight = round(1.5*size)
    holecentre = 0.5 + round(3*size)
    holeradius = round(2*size)
    holehighlight = round(1.5*size)
    shaftend = 0.5 + round(25*size)
    shaftwidth = round(2*size)
    shafthighlight = round(1.5*size)
    cmax = shaftend + shaftwidth

    # Define three line segments, such that the shortest distance
    # vectors from any point to each of these segments determines
    # everything we need to know about where it is on the spanner
    # shape.
    segments = [
    ((0,0), (holecentre, holecentre)),
    ((headcentre, headcentre), (headcentre, headcentre)),
    ((headcentre+headradius/math.sqrt(2), headcentre+headradius/math.sqrt(2)),
    (cmax, cmax))
    ]

    for y in range(int(cmax)):
        for x in range(int(cmax)):
            vectors = [linedist(a,b,c,d,x,y) for ((a,b),(c,d)) in segments]
            dists = [memoisedsqrt(vx*vx+vy*vy) for (vx,vy) in vectors]

            # If the distance to the hole line is less than
            # holeradius, we're not part of the spanner.
            if dists[0] < holeradius:
                continue
            # If the distance to the head `line' is less than
            # headradius, we are part of the spanner; likewise if
            # the distance to the shaft line is less than
            # shaftwidth _and_ the resulting shaft point isn't
            # beyond the shaft end.
            if dists[1] > headradius and \
            (dists[2] > shaftwidth or x+vectors[2][0] >= shaftend):
                continue

            # We're part of the spanner. Now compute the highlight
            # on this pixel. We do this by computing a `slope
            # vector', which points from this pixel in the
            # direction of its nearest edge. We store an array of
            # slope vectors, in polar coordinates.
            angles = [math.atan2(vy,vx) for (vx,vy) in vectors]
            slopes = []
            if dists[0] < holeradius + holehighlight:
                slopes.append(((dists[0]-holeradius)/holehighlight,angles[0]))
            if dists[1]/headradius < dists[2]/shaftwidth:
                if dists[1] > headradius - headhighlight and dists[1] < headradius:
                    slopes.append(((headradius-dists[1])/headhighlight,math.pi+angles[1]))
            else:
                if dists[2] > shaftwidth - shafthighlight and dists[2] < shaftwidth:
                    slopes.append(((shaftwidth-dists[2])/shafthighlight,math.pi+angles[2]))
            # Now we find the smallest distance in that array, if
            # any, and that gives us a notional position on a
            # sphere which we can use to compute the final
            # highlight level.
            bestdist = None
            bestangle = 0
            for dist, angle in slopes:
                if bestdist == None or bestdist > dist:
                    bestdist = dist
                    bestangle = angle
            if bestdist == None:
                bestdist = 1.0
            sx = (1.0-bestdist) * math.cos(bestangle)
            sy = (1.0-bestdist) * math.sin(bestangle)
            sz = math.sqrt(1.0 - sx*sx - sy*sy)
            shade = sx-sy+sz / math.sqrt(3) # can range from -1 to +1
            shade = 1.0 - (1-shade)/3

            pixel(x, y, yellowpix(shade), canvas)

    # And draw a border.
    border(canvas, size, [])

    return canvas

def box(size, back):
    canvas = {}

    # The back side of the cardboard box in the installer icon.

    boxwidth = round(15 * size)
    boxheight = round(12 * size)
    boxdepth = round(4 * size)
    boxfrontflapheight = round(5 * size)
    boxrightflapheight = round(3 * size)

    # Three shades of basically acceptable brown, all achieved by
    # halftoning between two of the Windows-16 colours. I'm quite
    # pleased that was feasible at all!
    dark = halftone(cr, cK)
    med = halftone(cr, cy)
    light = halftone(cr, cY)
    # We define our halftoning parity in such a way that the black
    # pixels along the RHS of the visible part of the box back
    # match up with the one-pixel black outline around the
    # right-hand side of the box. In other words, we want the pixel
    # at (-1, boxwidth-1) to be black, and hence the one at (0,
    # boxwidth) too.
    parityadjust = int(boxwidth) % 2

    # The entire back of the box.
    if back:
        for x in range(int(boxwidth + boxdepth)):
            ytop = max(-x-1, -boxdepth-1)
            ybot = min(boxheight, boxheight+boxwidth-1-x)
            for y in range(int(ytop), int(ybot)):
                pixel(x, y, dark[(x+y+parityadjust) % 2], canvas)

    # Even when drawing the back of the box, we still draw the
    # whole shape, because that means we get the right overall size
    # (the flaps make the box front larger than the box back) and
    # it'll all be overwritten anyway.

    # The front face of the box.
    for x in range(int(boxwidth)):
        for y in range(int(boxheight)):
            pixel(x, y, med[(x+y+parityadjust) % 2], canvas)
    # The right face of the box.
    for x in range(int(boxwidth), int(boxwidth+boxdepth)):
        ybot = boxheight + boxwidth-x
        ytop = ybot - boxheight
        for y in range(int(ytop), int(ybot)):
            pixel(x, y, dark[(x+y+parityadjust) % 2], canvas)
    # The front flap of the box.
    for y in range(int(boxfrontflapheight)):
        xadj = int(round(-0.5*y))
        for x in range(int(xadj), int(xadj+boxwidth)):
            pixel(x, y, light[(x+y+parityadjust) % 2], canvas)
    # The right flap of the box.
    for x in range(int(boxwidth), int(boxwidth + boxdepth + boxrightflapheight + 1)):
        ytop = max(boxwidth - 1 - x, x - boxwidth - 2*boxdepth - 1)
        ybot = min(x - boxwidth - 1, boxwidth + 2*boxrightflapheight - 1 - x)
        for y in range(int(ytop), int(ybot+1)):
            pixel(x, y, med[(x+y+parityadjust) % 2], canvas)

    # And draw a border.
    border(canvas, size, [(0, int(boxheight)-1, BL)])

    return canvas

def boxback(size):
    return box(size, 1)
def boxfront(size):
    return box(size, 0)

# Functions to draw entire icons by composing the above components.

def xybolt(c1, c2, size, boltoffx=0, boltoffy=0, aux={}):
    # Two unspecified objects and a lightning bolt.

    canvas = {}
    w = h = round(32 * size)

    bolt = lightning(size)

    # Position c2 against the top right of the icon.
    bb = bbox(c2)
    assert bb[2]-bb[0] <= w and bb[3]-bb[1] <= h
    overlay(c2, w-bb[2], 0-bb[1], canvas)
    aux["c2pos"] = (w-bb[2], 0-bb[1])
    # Position c1 against the bottom left of the icon.
    bb = bbox(c1)
    assert bb[2]-bb[0] <= w and bb[3]-bb[1] <= h
    overlay(c1, 0-bb[0], h-bb[3], canvas)
    aux["c1pos"] = (0-bb[0], h-bb[3])
    # Place the lightning bolt artistically off-centre. (The
    # rationale for this positioning is that it's centred on the
    # midpoint between the centres of the two monitors in the PuTTY
    # icon proper, but it's not really feasible to _base_ the
    # calculation here on that.)
    bb = bbox(bolt)
    assert bb[2]-bb[0] <= w and bb[3]-bb[1] <= h
    overlay(bolt, (w-bb[0]-bb[2])/2 + round(boltoffx*size), \
    (h-bb[1]-bb[3])/2 + round((boltoffy-2)*size), canvas)

    return canvas

def putty_icon(size):
    return xybolt(computer(size), computer(size), size)

def puttycfg_icon(size):
    w = h = round(32 * size)
    s = spanner(size)
    canvas = putty_icon(size)
    # Centre the spanner.
    bb = bbox(s)
    overlay(s, (w-bb[0]-bb[2])/2, (h-bb[1]-bb[3])/2, canvas)
    return canvas

def puttygen_icon(size):
    return xybolt(computer(size), key(size), size, boltoffx=2)

def pscp_icon(size):
    return xybolt(document(size), computer(size), size)

def puttyins_icon(size):
    aret = {}
    # The box back goes behind the lightning bolt.
    canvas = xybolt(boxback(size), computer(size), size, boltoffx=-2, boltoffy=+1, aux=aret)
    # But the box front goes over the top, so that the lightning
    # bolt appears to come _out_ of the box. Here it's useful to
    # know the exact coordinates where xybolt placed the box back,
    # so we can overlay the box front exactly on top of it.
    c1x, c1y = aret["c1pos"]
    overlay(boxfront(size), c1x, c1y, canvas)
    return canvas

def pterm_icon(size):
    # Just a really big computer.

    canvas = {}
    w = h = round(32 * size)

    c = computer(size * 1.4)

    # Centre c in the return canvas.
    bb = bbox(c)
    assert bb[2]-bb[0] <= w and bb[3]-bb[1] <= h
    overlay(c, (w-bb[0]-bb[2])/2, (h-bb[1]-bb[3])/2, canvas)

    return canvas

def ptermcfg_icon(size):
    w = h = round(32 * size)
    s = spanner(size)
    canvas = pterm_icon(size)
    # Centre the spanner.
    bb = bbox(s)
    overlay(s, (w-bb[0]-bb[2])/2, (h-bb[1]-bb[3])/2, canvas)
    return canvas

def pageant_icon(size):
    # A biggish computer, in a hat.

    canvas = {}
    w = h = round(32 * size)

    c = computer(size * 1.2)
    ht = hat(size)

    cbb = bbox(c)
    hbb = bbox(ht)

    # Determine the relative y-coordinates of the computer and hat.
    # We just centre the one on the other.
    xrel = (cbb[0]+cbb[2]-hbb[0]-hbb[2])/2

    # Determine the relative y-coordinates of the computer and hat.
    # We do this by sitting the hat as low down on the computer as
    # possible without any computer showing over the top. To do
    # this we first have to find the minimum x coordinate at each
    # y-coordinate of both components.
    cty = topy(c)
    hty = topy(ht)
    yrelmin = None
    for cx in cty.keys():
        hx = cx - xrel
        assert hty.has_key(hx)
        yrel = cty[cx] - hty[hx]
        if yrelmin == None:
            yrelmin = yrel
        else:
            yrelmin = min(yrelmin, yrel)

    # Overlay the hat on the computer.
    overlay(ht, xrel, yrelmin, c)

    # And centre the result in the main icon canvas.
    bb = bbox(c)
    assert bb[2]-bb[0] <= w and bb[3]-bb[1] <= h
    overlay(c, (w-bb[0]-bb[2])/2, (h-bb[1]-bb[3])/2, canvas)

    return canvas

# Test and output functions.

import os
import sys

def testrun(func, fname):
    canvases = []
    for size in [0.5, 0.6, 1.0, 1.2, 1.5, 4.0]:
        canvases.append(func(size))
    wid = 0
    ht = 0
    for canvas in canvases:
        minx, miny, maxx, maxy = bbox(canvas)
        wid = max(wid, maxx-minx+4)
        ht = ht + maxy-miny+4
    block = []
    for canvas in canvases:
        minx, miny, maxx, maxy = bbox(canvas)
        block.extend(render(canvas, minx-2, miny-2, minx-2+wid, maxy+2))
    p = os.popen("convert -depth 8 -size %dx%d rgb:- %s" % (wid,ht,fname), "w")
    assert len(block) == ht
    for line in block:
        assert len(line) == wid
        for r, g, b, a in line:
            # Composite on to orange.
            r = int(round((r * a + 255 * (255-a)) / 255.0))
            g = int(round((g * a + 128 * (255-a)) / 255.0))
            b = int(round((b * a +   0 * (255-a)) / 255.0))
            p.write("%c%c%c" % (r,g,b))
    p.close()

def drawicon(func, width, fname, orangebackground = 0):
    canvas = func(width / 32.0)
    finalise(canvas)
    minx, miny, maxx, maxy = bbox(canvas)
    assert minx >= 0 and miny >= 0 and maxx <= width and maxy <= width

    block = render(canvas, 0, 0, width, width)
    p = os.popen("convert -depth 8 -size %dx%d rgba:- %s" % (width,width,fname), "w")
    assert len(block) == width
    for line in block:
        assert len(line) == width
        for r, g, b, a in line:
            if orangebackground:
                # Composite on to orange.
                r = int(round((r * a + 255 * (255-a)) / 255.0))
                g = int(round((g * a + 128 * (255-a)) / 255.0))
                b = int(round((b * a +   0 * (255-a)) / 255.0))
                a = 255
            p.write("%c%c%c%c" % (r,g,b,a))
    p.close()

args = sys.argv[1:]

orangebackground = test = 0
colours = 1 # 0=mono, 1=16col, 2=truecol
doingargs = 1

realargs = []
for arg in args:
    if doingargs and arg[0] == "-":
        if arg == "-t":
            test = 1
        elif arg == "-it":
            orangebackground = 1
        elif arg == "-2":
            colours = 0
        elif arg == "-T":
            colours = 2
        elif arg == "--":
            doingargs = 0
        else:
            sys.stderr.write("unrecognised option '%s'\n" % arg)
            sys.exit(1)
    else:
        realargs.append(arg)

if colours == 0:
    # Monochrome.
    cK=cr=cg=cb=cm=cc=cP=cw=cR=cG=cB=cM=cC=cD = 0
    cY=cy=cW = 1
    cT = -1
    def greypix(value):
        return [cK,cW][int(round(value))]
    def yellowpix(value):
        return [cK,cW][int(round(value))]
    def bluepix(value):
        return cK
    def dark(value):
        return [cT,cK][int(round(value))]
    def blend(col1, col2):
        if col1 == cT:
            return col2
        else:
            return col1
    pixvals = [
    (0x00, 0x00, 0x00, 0xFF), # cK
    (0xFF, 0xFF, 0xFF, 0xFF), # cW
    (0x00, 0x00, 0x00, 0x00), # cT
    ]
    def outpix(colour):
        return pixvals[colour]
    def finalisepix(colour):
        return colour
    def halftone(col1, col2):
        return (col1, col2)
elif colours == 1:
    # Windows 16-colour palette.
    cK,cr,cg,cy,cb,cm,cc,cP,cw,cR,cG,cY,cB,cM,cC,cW = range(16)
    cT = -1
    cD = -2 # special translucent half-darkening value used internally
    def greypix(value):
        return [cK,cw,cw,cP,cW][int(round(4*value))]
    def yellowpix(value):
        return [cK,cy,cY][int(round(2*value))]
    def bluepix(value):
        return [cK,cb,cB][int(round(2*value))]
    def dark(value):
        return [cT,cD,cK][int(round(2*value))]
    def blend(col1, col2):
        if col1 == cT:
            return col2
        elif col1 == cD:
            return [cK,cK,cK,cK,cK,cK,cK,cw,cK,cr,cg,cy,cb,cm,cc,cw,cD,cD][col2]
        else:
            return col1
    pixvals = [
    (0x00, 0x00, 0x00, 0xFF), # cK
    (0x80, 0x00, 0x00, 0xFF), # cr
    (0x00, 0x80, 0x00, 0xFF), # cg
    (0x80, 0x80, 0x00, 0xFF), # cy
    (0x00, 0x00, 0x80, 0xFF), # cb
    (0x80, 0x00, 0x80, 0xFF), # cm
    (0x00, 0x80, 0x80, 0xFF), # cc
    (0xC0, 0xC0, 0xC0, 0xFF), # cP
    (0x80, 0x80, 0x80, 0xFF), # cw
    (0xFF, 0x00, 0x00, 0xFF), # cR
    (0x00, 0xFF, 0x00, 0xFF), # cG
    (0xFF, 0xFF, 0x00, 0xFF), # cY
    (0x00, 0x00, 0xFF, 0xFF), # cB
    (0xFF, 0x00, 0xFF, 0xFF), # cM
    (0x00, 0xFF, 0xFF, 0xFF), # cC
    (0xFF, 0xFF, 0xFF, 0xFF), # cW
    (0x00, 0x00, 0x00, 0x80), # cD
    (0x00, 0x00, 0x00, 0x00), # cT
    ]
    def outpix(colour):
        return pixvals[colour]
    def finalisepix(colour):
        # cD is used internally, but can't be output. Convert to cK.
        if colour == cD:
            return cK
        return colour
    def halftone(col1, col2):
        return (col1, col2)
else:
    # True colour.
    cK = (0x00, 0x00, 0x00, 0xFF)
    cr = (0x80, 0x00, 0x00, 0xFF)
    cg = (0x00, 0x80, 0x00, 0xFF)
    cy = (0x80, 0x80, 0x00, 0xFF)
    cb = (0x00, 0x00, 0x80, 0xFF)
    cm = (0x80, 0x00, 0x80, 0xFF)
    cc = (0x00, 0x80, 0x80, 0xFF)
    cP = (0xC0, 0xC0, 0xC0, 0xFF)
    cw = (0x80, 0x80, 0x80, 0xFF)
    cR = (0xFF, 0x00, 0x00, 0xFF)
    cG = (0x00, 0xFF, 0x00, 0xFF)
    cY = (0xFF, 0xFF, 0x00, 0xFF)
    cB = (0x00, 0x00, 0xFF, 0xFF)
    cM = (0xFF, 0x00, 0xFF, 0xFF)
    cC = (0x00, 0xFF, 0xFF, 0xFF)
    cW = (0xFF, 0xFF, 0xFF, 0xFF)
    cD = (0x00, 0x00, 0x00, 0x80)
    cT = (0x00, 0x00, 0x00, 0x00)
    def greypix(value):
        value = max(min(value, 1), 0)
        return (int(round(0xFF*value)),) * 3 + (0xFF,)
    def yellowpix(value):
        value = max(min(value, 1), 0)
        return (int(round(0xFF*value)),) * 2 + (0, 0xFF)
    def bluepix(value):
        value = max(min(value, 1), 0)
        return (0, 0, int(round(0xFF*value)), 0xFF)
    def dark(value):
        value = max(min(value, 1), 0)
        return (0, 0, 0, int(round(0xFF*value)))
    def blend(col1, col2):
        r1,g1,b1,a1 = col1
        r2,g2,b2,a2 = col2
        r = int(round((r1*a1 + r2*(0xFF-a1)) / 255.0))
        g = int(round((g1*a1 + g2*(0xFF-a1)) / 255.0))
        b = int(round((b1*a1 + b2*(0xFF-a1)) / 255.0))
        a = int(round((255*a1 + a2*(0xFF-a1)) / 255.0))
        return r, g, b, a
    def outpix(colour):
        return colour
    if colours == 2:
        # True colour with no alpha blending: we still have to
        # finalise half-dark pixels to black.
        def finalisepix(colour):
            if colour[3] > 0:
                return colour[:3] + (0xFF,)
            return colour
    else:
        def finalisepix(colour):
            return colour
    def halftone(col1, col2):
        r1,g1,b1,a1 = col1
        r2,g2,b2,a2 = col2
        colret = (int(r1+r2)/2, int(g1+g2)/2, int(b1+b2)/2, int(a1+a2)/2)
        return (colret, colret)

if test:
    testrun(eval(realargs[0]), realargs[1])
else:
    drawicon(eval(realargs[0]), int(realargs[1]), realargs[2], orangebackground)



# -*- coding: utf-8 -*-
#
# AWL simulator - Code validator
#
# Copyright 2014-2016 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.common.util import *
from awlsim.common.exceptions import *

from awlsim.coreclient.client import *

import threading


__all__ = [
	"AwlValidator",
]


class AwlValidator(object):
	"""Source code validation.
	"""

	__globalInstance = None

	_PORT_RANGE	= range(30000, 32000)
	_EXIT_THREAD	= -1

	@classmethod
	def startup(cls):
		if cls.__globalInstance:
			return cls.__globalInstance
		printDebug("AwlValidator: Startup")
		try:
			val = cls()
		except AwlSimError as e:
			printDebug("Failed to startup validator: %s" % str(e))
			return None
		cls.__globalInstance = val
		return val

	@classmethod
	def shutdown(cls):
		if cls.__globalInstance:
			cls.__globalInstance.doShutdown()
			cls.__globalInstance = None

	@classmethod
	def get(cls):
		return cls.startup()

	def __init__(self):
		self.__thread = None
		self.__job = None
		self.__exception = None
		self.__running = False
		self.__lock = threading.Lock()
		self.__condition = threading.Condition(self.__lock)

		self.__client = AwlSimClient()
		try:
			self.__client.setLoglevel(Logging.LOG_ERROR,
						  setClientSide=False,
						  setServerSide=True)
			self.__client.spawnServer(listenHost = "localhost",
						  listenPort = self._PORT_RANGE)
			self.__client.connectToServer(host = "localhost",
						      port = self.__client.serverProcessPort)
			self.__client.setLoglevel(Logging.LOG_NONE,
						  setClientSide=False,
						  setServerSide=True)
		except AwlSimError as e:
			self.doShutdown()
			raise e

		self.__thread = threading.Thread(target=self.__threadFunc)
		self.__thread.daemon = False
		self.__thread.start()

	def doShutdown(self):
		printDebug("AwlValidator: Shutdown")
		if self.__client:
			self.__client.shutdown()
			self.__client = None
		if self.__thread:
			with self.__lock:
				self.__job = self._EXIT_THREAD
				self.__condition.notify_all()
			self.__thread.join()
			self.__thread = None
			self.__exception = None
			self.__running = False

	def __runJob(self, job):
		(project,
		 symTabSources, libSelections, awlSources,
		 fupSources, kopSources) = job

		client = self.__client
		exception = None
		try:
			client.setRunState(False)
			client.reset()
			client.loadProject(project, loadSymTabs=False,
					   loadLibSelections=False,
					   loadSources=False,
					   loadFup=False,
					   loadKop=False)
			client.loadSymTabSources(symTabSources)
			client.loadLibraryBlocks(libSelections)
			client.loadAwlSources(awlSources)
			client.loadFupSources(fupSources)
			client.loadKopSources(kopSources)
			client.build()
			client.reset()
		except AwlSimError as e:
			exception = e
		with self.__lock:
			if self.__job is None:
				self.__running = False
			self.__exception = exception

	def __threadFunc(self):
		"""This is the validation thread.
		"""
		while True:
			with self.__lock:
				if self.__job is None:
					self.__condition.wait()
				job, self.__job = self.__job, None
			if job is self._EXIT_THREAD:
				break
			if job:
				self.__runJob(job)

	def validate(self, project,
		     symTabSources, libSelections, awlSources,
		     fupSources, kopSources):
		"""Schedule a validation.
		Get the result with getState().
		"""
		if not project or not awlSources:
			return
		with self.__lock:
			if self.__job is self._EXIT_THREAD:
				return
			self.__job = (project,
				      symTabSources, libSelections, awlSources,
				      fupSources, kopSources)
			self.__running = True
			self.__condition.notify_all()

	def getState(self):
		"""Get the validation result.
		Returns a tuple (running, exception).
		"""
		with self.__lock:
			running = self.__running
			exception = self.__exception
		return running, exception



from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim_tstlib import *

from awlsim.gui.geo2d import *


class Test_LineSeg2D(object):
	def test_point(self):
		p = Point2D()
		assert_eq(p.x, 0)
		assert_eq(p.y, 0)
		assert_eq(bool(p), False)

		p = Point2D(42, -42)
		assert_eq(p.x, 42)
		assert_eq(p.y, -42)
		assert_eq(p, Point2D(42, -42))
		assert_ne(p, Point2D(42, 42))
		assert_ne(p, Point2D(-42, -42))
		assert_eq(bool(p), True)

	def test_vect(self):
		v = Vect2D()
		assert_eq(v.x, 0)
		assert_eq(v.y, 0)
		assert_eq(bool(v), False)

		v = Vect2D(42, -42)
		assert_eq(v.x, 42)
		assert_eq(v.y, -42)
		assert_eq(v, Vect2D(42, -42))
		assert_ne(v, Vect2D(42, 42))
		assert_ne(v, Vect2D(-42, -42))
		assert_eq(bool(v), True)

	def test_lineseg(self):
		s = LineSeg2D.fromCoords(3, 4, 3, 4)
		assert_eq(s.pointA, Point2D(3, 4))
		assert_eq(s.pointB, Point2D(3, 4))
		assert_eq(bool(s), False)

		s = LineSeg2D.fromCoords(1, 2, 3, 4)
		assert_eq(s.pointA, Point2D(1, 2))
		assert_eq(s.pointB, Point2D(3, 4))
		assert_eq(bool(s), True)

		assert_eq(LineSeg2D.fromCoords(1, 2, 3, 4), LineSeg2D.fromCoords(1, 2, 3, 4))
		assert_ne(LineSeg2D.fromCoords(1, 2, 3, 4), LineSeg2D.fromCoords(-1, 2, 3, 4))
		assert_ne(LineSeg2D.fromCoords(1, 2, 3, 4), LineSeg2D.fromCoords(1, -2, 3, 4))
		assert_ne(LineSeg2D.fromCoords(1, 2, 3, 4), LineSeg2D.fromCoords(1, 2, -3, 4))
		assert_ne(LineSeg2D.fromCoords(1, 2, 3, 4), LineSeg2D.fromCoords(1, 2, 3, -4))
		assert_eq(bool(s), True)

	def test_intersection_base(self):
		inter = Inter2D()
		assert_eq(inter, Inter2D())
		assert_ne(inter, Inter2D(Point2D(1, 1)))
		assert_is(inter.intersects, False)
		assert_is(inter.point, None)
		assert_eq(inter.vect, Vect2D())
		assert_eq(bool(inter), False)

		inter = Inter2D(Point2D(42, -42), Vect2D(142, -142), False)
		assert_eq(inter, Inter2D(Point2D(42, -42), Vect2D(142, -142), False))
		assert_ne(inter, Inter2D(Point2D(42, -42), Vect2D(143, -142), False))
		assert_is(inter.intersects, False)
		assert_eq(inter.point, Point2D(42, -42))
		assert_eq(inter.vect, Vect2D(142, -142))
		assert_eq(bool(inter), False)

		inter = Inter2D(Point2D(42, -42), Vect2D(142, -142), True)
		assert_eq(inter, Inter2D(Point2D(42, -42), Vect2D(142, -142), True))
		assert_ne(inter, Inter2D(Point2D(42, -42), Vect2D(142, -142), False))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(42, -42))
		assert_eq(inter.vect, Vect2D(142, -142))
		assert_eq(bool(inter), True)

		inter = Inter2D(None, Vect2D(142, -142), True)
		assert_is(inter.intersects, False)
		assert_is(inter.point, None)
		assert_eq(inter.vect, Vect2D(142, -142))
		assert_eq(bool(inter), False)

	def test_intersection(self):
		# intersecting
		inter = LineSeg2D(Point2D(2, 7), Point2D(8, 1)).intersection(
			LineSeg2D(Point2D(9, 8), Point2D(3, 2)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(5, 4))
		assert_eq(inter.vect, Vect2D())
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(5, 4), Point2D(5, 4)))

		# not intersecting
		inter = LineSeg2D(Point2D(1, 1), Point2D(4, 5)).intersection(
			LineSeg2D(Point2D(6, 6), Point2D(8, 2)))
		assert_is(inter.intersects, False)
		assert_eq(inter.point, Point2D(5.5, 7))
		assert_eq(inter.vect, Vect2D())
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(5.5, 7), Point2D(5.5, 7)))

		# parallel, horizontal, not intersecting
		inter = LineSeg2D(Point2D(3, 2), Point2D(9, 2)).intersection(
			LineSeg2D(Point2D(5, 3), Point2D(14, 3)))
		assert_is(inter.intersects, False)
		assert_is_none(inter.point)
		assert_eq(inter.vect, Vect2D())
		assert_is_none(inter.lineSeg)

		# parallel, vertical, not intersecting
		inter = LineSeg2D(Point2D(3, 9), Point2D(3, 4)).intersection(
			LineSeg2D(Point2D(5, 2), Point2D(5, 5)))
		assert_is(inter.intersects, False)
		assert_is_none(inter.point)
		assert_eq(inter.vect, Vect2D())
		assert_is_none(inter.lineSeg)

		# parallel, horizontal, intersecting
		inter = LineSeg2D(Point2D(3, 2), Point2D(9, 2)).intersection(
			LineSeg2D(Point2D(5, 2), Point2D(14, 2)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(9, 2))
		assert_eq(inter.vect, Vect2D(-4, 0))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(9, 2), Point2D(5, 2)))

		# parallel, horizontal, intersecting
		inter = LineSeg2D(Point2D(3, 2), Point2D(9, 2)).intersection(
			LineSeg2D(Point2D(14, 2), Point2D(5, 2)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(9, 2))
		assert_eq(inter.vect, Vect2D(-4, 0))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(9, 2), Point2D(5, 2)))

		# parallel, horizontal, intersecting
		inter = LineSeg2D(Point2D(9, 2), Point2D(3, 2)).intersection(
			LineSeg2D(Point2D(14, 2), Point2D(5, 2)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(9, 2))
		assert_eq(inter.vect, Vect2D(-4, 0))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(9, 2), Point2D(5, 2)))

		# parallel, horizontal, intersecting
		inter = LineSeg2D(Point2D(5, 2), Point2D(14, 2)).intersection(
			LineSeg2D(Point2D(3, 2), Point2D(9, 2)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(5, 2))
		assert_eq(inter.vect, Vect2D(4, 0))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(5, 2), Point2D(9, 2)))

		# parallel, vertical, intersecting
		inter = LineSeg2D(Point2D(3, 9), Point2D(3, 4)).intersection(
			LineSeg2D(Point2D(3, 2), Point2D(3, 5)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(3, 4))
		assert_eq(inter.vect, Vect2D(0, 1))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(3, 4), Point2D(3, 5)))

		# parallel, vertical, intersecting
		inter = LineSeg2D(Point2D(3, 2), Point2D(3, 5)).intersection(
			LineSeg2D(Point2D(3, 9), Point2D(3, 4)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(3, 5))
		assert_eq(inter.vect, Vect2D(0, -1))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(3, 5), Point2D(3, 4)))

		# parallel, vertical, intersecting
		inter = LineSeg2D(Point2D(60, 30), Point2D(60, 90)).intersection(
			LineSeg2D(Point2D(60, 30), Point2D(60, 70)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(60, 30))
		assert_eq(inter.vect, Vect2D(0, 40))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(60, 30), Point2D(60, 70)))

		# parallel, vertical, intersecting
		inter = LineSeg2D(Point2D(4, 8), Point2D(4, 2)).intersection(
			LineSeg2D(Point2D(4, 8), Point2D(4, 4)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 8))
		assert_eq(inter.vect, Vect2D(0, -4))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 8), Point2D(4, 4)))

		# parallel, vertical, intersecting
		inter = LineSeg2D(Point2D(4, 8), Point2D(4, 2)).intersection(
			LineSeg2D(Point2D(4, 6), Point2D(4, 2)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 2))
		assert_eq(inter.vect, Vect2D(0, 4))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 2), Point2D(4, 6)))

		# parallel, vertical, intersecting
		inter = LineSeg2D(Point2D(4, 8), Point2D(4, 2)).intersection(
			LineSeg2D(Point2D(4, 7), Point2D(4, 3)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 7))
		assert_eq(inter.vect, Vect2D(0, -4))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 7), Point2D(4, 3)))

		# parallel, vertical, intersecting
		inter = LineSeg2D(Point2D(4, 8), Point2D(4, 4)).intersection(
			LineSeg2D(Point2D(4, 8), Point2D(4, 2)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 8))
		assert_eq(inter.vect, Vect2D(0, -4))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 8), Point2D(4, 4)))

		# parallel, vertical, intersecting
		inter = LineSeg2D(Point2D(4, 6), Point2D(4, 2)).intersection(
			LineSeg2D(Point2D(4, 8), Point2D(4, 2)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 6))
		assert_eq(inter.vect, Vect2D(0, -4))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 6), Point2D(4, 2)))

		# parallel, vertical, intersecting
		inter = LineSeg2D(Point2D(4, 7), Point2D(4, 3)).intersection(
			LineSeg2D(Point2D(4, 8), Point2D(4, 2)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 7))
		assert_eq(inter.vect, Vect2D(0, -4))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 7), Point2D(4, 3)))

		# parallel, vertical, pointy intersection
		inter = LineSeg2D(Point2D(4, 1), Point2D(4, 5)).intersection(
			LineSeg2D(Point2D(4, 5), Point2D(4, 7)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 5))
		assert_eq(inter.vect, Vect2D(0, 0))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 5), Point2D(4, 5)))

		# parallel, vertical, pointy intersection
		inter = LineSeg2D(Point2D(4, 5), Point2D(4, 1)).intersection(
			LineSeg2D(Point2D(4, 5), Point2D(4, 7)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 5))
		assert_eq(inter.vect, Vect2D(0, 0))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 5), Point2D(4, 5)))

		# parallel, vertical, pointy intersection
		inter = LineSeg2D(Point2D(4, 5), Point2D(4, 1)).intersection(
			LineSeg2D(Point2D(4, 7), Point2D(4, 5)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 5))
		assert_eq(inter.vect, Vect2D(0, 0))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 5), Point2D(4, 5)))

		# parallel, vertical, pointy intersection
		inter = LineSeg2D(Point2D(4, 1), Point2D(4, 5)).intersection(
			LineSeg2D(Point2D(4, 7), Point2D(4, 5)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 5))
		assert_eq(inter.vect, Vect2D(0, 0))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 5), Point2D(4, 5)))

		# parallel, vertical, full intersection
		inter = LineSeg2D(Point2D(4, 1), Point2D(4, 7)).intersection(
			LineSeg2D(Point2D(4, 1), Point2D(4, 7)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 1))
		assert_eq(inter.vect, Vect2D(0, 6))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 1), Point2D(4, 7)))

		# parallel, vertical, full intersection
		inter = LineSeg2D(Point2D(4, 7), Point2D(4, 1)).intersection(
			LineSeg2D(Point2D(4, 7), Point2D(4, 1)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 7))
		assert_eq(inter.vect, Vect2D(0, -6))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 7), Point2D(4, 1)))

		# parallel, vertical, full intersection
		inter = LineSeg2D(Point2D(4, 1), Point2D(4, 7)).intersection(
			LineSeg2D(Point2D(4, 7), Point2D(4, 1)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 1))
		assert_eq(inter.vect, Vect2D(0, 6))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 1), Point2D(4, 7)))

		# parallel, vertical, full intersection
		inter = LineSeg2D(Point2D(4, 7), Point2D(4, 1)).intersection(
			LineSeg2D(Point2D(4, 1), Point2D(4, 7)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 7))
		assert_eq(inter.vect, Vect2D(0, -6))
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 7), Point2D(4, 1)))

		# AB vertical, CD horizontal, intersecting
		inter = LineSeg2D(Point2D(4, 1), Point2D(4, 8)).intersection(
			LineSeg2D(Point2D(2, 5), Point2D(12, 5)))
		assert_is(inter.intersects, True)
		assert_eq(inter.point, Point2D(4, 5))
		assert_eq(inter.vect, Vect2D())
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(4, 5), Point2D(4, 5)))

		# AB vertical, CD horizontal, not intersecting
		inter = LineSeg2D(Point2D(x=180, y=130), Point2D(x=180, y=110)).intersection(
			LineSeg2D(Point2D(x=236, y=110), Point2D(x=240, y=110)))
		assert_is(inter.intersects, False)
		assert_eq(inter.point, Point2D(180, 110))
		assert_eq(inter.vect, Vect2D())
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(180, 110), Point2D(180, 110)))

		# AB horizontal, CD vertical, not intersecting
		inter = LineSeg2D(Point2D(9, 5), Point2D(4, 5)).intersection(
			LineSeg2D(Point2D(8, 6), Point2D(8, 11)))
		assert_is(inter.intersects, False)
		assert_eq(inter.point, Point2D(8, 5))
		assert_eq(inter.vect, Vect2D())
		assert_eq(inter.lineSeg, LineSeg2D(Point2D(8, 5), Point2D(8, 5)))
